# Copyright 2004-2005 Joe Wreschnig, Michael Urman, Iñigo Serna
#                2013 Christoph Reiter
#           2020-2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import time
from collections.abc import Collection

from gi.repository import GObject, GLib

from quodlibet import config, print_d
from quodlibet.formats import AudioFile
from quodlibet.library.base import Library


class TimeTracker(GObject.GObject):
    """Emits tick every second (with up to one second jitter) as long
    as the player is actively playing.

    Uses timeout_add_seconds, so multiple instances of this should
    sync and not produce more wakeups.
    """

    __gsignals__ = {
        "tick": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, player):
        super().__init__()

        self.__interval = 1000
        self.__player = player
        self.__id = None
        self.__stop = False
        self.__reset = False
        self.__sigs = [
            player.connect("paused", self.__paused, True),
            player.connect("unpaused", self.__paused, False),
        ]
        self.__paused(player, player.paused)

    def set_interval(self, interval):
        """Update the resolution in milliseconds"""

        self.__interval = interval
        self.__reset = True

    def tick(self):
        """Emit a tick event"""

        self.emit("tick")

    def destroy(self):
        for signal_id in self.__sigs:
            self.__player.disconnect(signal_id)
        self.__source_remove()

    def __source_remove(self):
        if self.__id is not None:
            GLib.source_remove(self.__id)
            self.__id = None

    def __update(self):
        if self.__stop:
            self.__source_remove()
            return False

        if self.__reset:
            self.__reset = False
            self.__source_remove()
            self.__paused(self.__player, self.__player.paused)

        self.tick()
        return True

    def __paused(self, player, paused):
        if paused:
            # By removing the timeout only in the callback we are safe from
            # huge deviation caused by lots of pause/unpause actions.
            self.__stop = True
        else:
            self.__stop = False
            if self.__id is None:
                # The application is already woke up every seconds
                # so synchronize to it by calling timeout_add_seconds(...)
                # if the requested tracker interval is exactly 1 second.
                self.__id = (
                    GLib.timeout_add_seconds(1, self.__update)
                    if self.__interval == 1000
                    else GLib.timeout_add(self.__interval, self.__update)
                )


class SongTracker:
    def __init__(self, librarian, player, pl):
        self.__player_ids = [
            player.connect("song-ended", self.__end, librarian, pl),
            player.connect("song-started", self.__start, librarian),
        ]
        self.__player = player
        timer = TimeTracker(player)
        timer.connect("tick", self.__timer)
        self.elapsed = 0
        self.__to_change = set()
        self.__change_id = None

    def destroy(self):
        for id_ in self.__player_ids:
            self.__player.disconnect(id_)
        self.__player = None

        if self.__change_id:
            GLib.source_remove(self.__change_id)
            self.__change_id = None

    def __changed(self, librarian, song):
        # try to combine changed events and process them if QL is idle
        self.__to_change.add(song)

        if self.__change_id is not None:
            GLib.source_remove(self.__change_id)
            self.__change_id = None

        def idle_change():
            librarian.changed(list(self.__to_change))
            self.__to_change.clear()
            self.__change_id = None

        self.__change_id = GLib.idle_add(idle_change, priority=GLib.PRIORITY_LOW)

    def __start(self, player, song, librarian):
        self.elapsed = 0
        if song is not None:
            if song.multisong:
                song["~#lastplayed"] = int(time.time())
                song["~#playcount"] = song.get("~#playcount", 0) + 1
            else:
                config.set("memory", "song", song["~filename"])
            song["~#laststarted"] = int(time.time())
            self.__changed(librarian, song)
        else:
            config.set("memory", "song", "")

    def __end(self, player, song, ended, librarian, pl):
        if song is not None and not song.multisong:
            if ended:
                config.set("memory", "seek", player.get_position())
            else:
                config.set("memory", "seek", 0)

            playcount_minimum_length = config.getfloat(
                "player", "playcount_minimum_length_proportion"
            ) * int(song.get("~#length", 1))

            if self.elapsed >= playcount_minimum_length:
                song["~#lastplayed"] = int(time.time())
                song["~#playcount"] = song.get("~#playcount", 0) + 1
                self.__changed(librarian, song)
            elif pl.current is not song:
                if not player.error:
                    song["~#skipcount"] = song.get("~#skipcount", 0) + 1
                    self.__changed(librarian, song)
        else:
            config.set("memory", "seek", 0)

    def __timer(self, timer):
        self.elapsed += 1


class FSInterface:
    """Provides a file in ~/.quodlibet to indicate what song is playing."""

    def __init__(self, path, player, library: Library):
        self.path = path
        self._player = player
        self._pids = [
            player.connect("song-started", self.__started),
            player.connect("song-ended", self.__ended),
        ]
        self._lids = [library.connect("changed", self.__changed)]
        self._library = library

    def destroy(self):
        for id_ in self._pids:
            self._player.disconnect(id_)
        for id_ in self._lids:
            print_d(f"Disconnecting signal {id_} from {self._library}")
            self._library.disconnect(id_)
        try:
            os.unlink(self.path)
        except OSError:
            pass

    def __started(self, player, song):
        if song:
            song = player.with_elapsed_info(song)
            try:
                with open(self.path, "wb") as f:
                    f.write(song.to_dump())
            except OSError:
                pass

    def __ended(self, player, song, stopped):
        try:
            os.unlink(self.path)
        except OSError:
            pass

    def __changed(self, _lib: Library, songs: Collection[AudioFile]):
        current = self._player.song
        if current and current in songs:
            print_d("Current song changed, updating current file")
            self.__started(self._player, current)



# Copyright 2012 Christoph Reiter
#           2020 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import warnings
import logging

from senf import fsn2text

from quodlibet.const import MinVersions
from quodlibet import config
from quodlibet.util import is_osx, is_windows, i18n
from quodlibet.util.dprint import print_e, PrintHandler, print_d
from quodlibet.util.urllib import install_urllib2_ca_file

from ._main import get_base_dir, is_release, get_image_dir, get_cache_dir


_cli_initialized = False
_initialized = False


def _init_gtk_debug(no_excepthook):
    from quodlibet.errorreport import enable_errorhook

    enable_errorhook(not no_excepthook)


def is_init():
    """Returns if init() was called"""

    global _initialized

    return _initialized


def init(no_translations=False, no_excepthook=False, config_file=None):
    """This needs to be called before any API can be used.
    Might raise in case of an error.

    Pass no_translations=True to disable translations (used by tests)
    """

    global _initialized

    if _initialized:
        return

    init_cli(no_translations=no_translations, config_file=config_file)
    _init_gtk()
    _init_gtk_debug(no_excepthook=no_excepthook)
    _init_gst()
    _init_dbus()

    _initialized = True


def _init_gettext(no_translations=False):
    """Call before using gettext helpers"""

    if no_translations:
        language = "C"
    else:
        language = config.gettext("settings", "language")
        if language:
            print_d(f"Using language in QL settings: {language!r}")
        else:
            language = None

    i18n.init(language)

    # Use the locale dir in ../build/share/locale if there is one
    localedir = os.path.join(
        os.path.dirname(get_base_dir()), "build", "share", "locale"
    )
    if os.path.isdir(localedir):
        print_d(f"Using local locale dir {localedir}")
    else:
        localedir = None

    i18n.register_translation("quodlibet", localedir)
    debug_text = os.environ.get("QUODLIBET_TEST_TRANS")
    if debug_text is not None:
        i18n.set_debug_text(fsn2text(debug_text))


def _init_python():
    MinVersions.PYTHON3.check(sys.version_info)

    if is_osx():
        # We build our own openssl on OSX and need to make sure that
        # our own ca file is used in all cases as the non-system openssl
        # doesn't use the system certs
        install_urllib2_ca_file()

    if is_windows():
        # Not really needed on Windows as pygi-aio seems to work fine, but
        # wine doesn't have certs which we use for testing.
        install_urllib2_ca_file()

    if is_windows() and os.sep != "\\":
        # In the MSYS2 console MSYSTEM is set, which breaks os.sep/os.path.sep
        # If you hit this, do a "setup.py clean -all" to get rid of the
        # bytecode cache then start things with "MSYSTEM= ..."
        raise AssertionError("MSYSTEM is set ({!r})".format(os.environ.get("MSYSTEM")))

    logging.getLogger().addHandler(PrintHandler())


def _init_formats():
    from quodlibet.formats import init

    init()


def init_cli(no_translations=False, config_file=None):
    """This needs to be called before any API can be used.
    Might raise in case of an error.

    Like init() but for code not using Gtk etc.
    """

    global _cli_initialized

    if _cli_initialized:
        return

    _init_python()
    config.init_defaults()
    if config_file is not None:
        config.init(config_file)
    _init_gettext(no_translations)
    _init_formats()
    _init_g()

    _cli_initialized = True


def _init_dbus():
    """Setup dbus mainloop integration. Call before using dbus"""

    # To make GDBus fail early and we don't have to wait for a timeout
    if is_osx() or is_windows():
        os.environ["DBUS_SYSTEM_BUS_ADDRESS"] = "something-invalid"
        os.environ["DBUS_SESSION_BUS_ADDRESS"] = "something-invalid"

    try:
        from dbus.mainloop.glib import DBusGMainLoop, threads_init
    except ImportError:
        try:
            import dbus.glib

            dbus.glib  # noqa
        except ImportError:
            return
    else:
        threads_init()
        DBusGMainLoop(set_as_default=True)


def _fix_gst_leaks():
    """gst_element_add_pad and gst_bin_add are wrongly annotated and lead
    to PyGObject refing the passed element.

    Work around by adding a wrapper that unrefs afterwards.
    Can be called multiple times.

    https://bugzilla.gnome.org/show_bug.cgi?id=741390
    https://bugzilla.gnome.org/show_bug.cgi?id=702960
    """

    from gi.repository import Gst

    assert Gst.is_initialized()

    def do_wrap(func):
        def wrap(self, obj):
            result = func(self, obj)
            obj.unref()
            return result

        return wrap

    parent = Gst.Bin()
    elm = Gst.Bin()
    parent.add(elm)
    if elm.__grefcount__ == 3:
        elm.unref()
        Gst.Bin.add = do_wrap(Gst.Bin.add)

    pad = Gst.Pad.new("foo", Gst.PadDirection.SRC)
    parent.add_pad(pad)
    if pad.__grefcount__ == 3:
        pad.unref()
        Gst.Element.add_pad = do_wrap(Gst.Element.add_pad)


def _init_g():
    """Call before using GdkPixbuf/GLib/Gio/GObject"""

    import gi

    gi.require_version("GLib", "2.0")
    gi.require_version("Gio", "2.0")
    gi.require_version("GObject", "2.0")
    gi.require_version("GdkPixbuf", "2.0")

    # Newer glib is noisy regarding deprecated signals/properties
    # even with stable releases.
    if is_release():
        warnings.filterwarnings(
            "ignore", ".* It will be removed in a future version.", Warning
        )

    # blacklist some modules, simply loading can cause segfaults
    sys.modules["glib"] = None
    sys.modules["gobject"] = None


def _init_gtk():
    """Call before using Gtk/Gdk"""

    import gi

    if (
        config.getboolean("settings", "pangocairo_force_fontconfig")
        and "PANGOCAIRO_BACKEND" not in os.environ
    ):
        os.environ["PANGOCAIRO_BACKEND"] = "fontconfig"

    # disable for consistency and trigger events seem a bit flaky here
    if config.getboolean("settings", "scrollbar_always_visible"):
        os.environ["GTK_OVERLAY_SCROLLING"] = "0"

    try:
        # not sure if this is available under Windows
        gi.require_version("GdkX11", "3.0")
        from gi.repository import GdkX11

        GdkX11  # noqa
    except (ValueError, ImportError):
        pass

    gi.require_version("Gtk", "3.0")
    gi.require_version("Gdk", "3.0")
    gi.require_version("Pango", "1.0")
    gi.require_version("Soup", "3.0")
    gi.require_version("PangoCairo", "1.0")

    from gi.repository import Gtk
    from quodlibet.qltk import ThemeOverrider, gtk_version

    # PyGObject doesn't fail anymore when init fails, so do it ourselves
    initialized, sys.argv[:] = Gtk.init_check(sys.argv)
    if not initialized:
        raise SystemExit("Gtk.init failed")

    # include our own icon theme directory
    theme = Gtk.IconTheme.get_default()
    theme_search_path = get_image_dir()
    assert os.path.exists(theme_search_path)
    theme.append_search_path(theme_search_path)

    # Force menu/button image related settings. We might show too many atm
    # but this makes sure we don't miss cases where we forgot to force them
    # per widget.
    # https://bugzilla.gnome.org/show_bug.cgi?id=708676
    warnings.filterwarnings("ignore", ".*g_value_get_int.*", Warning)

    # some day... but not now
    warnings.filterwarnings("ignore", ".*Stock items are deprecated.*", Warning)
    warnings.filterwarnings("ignore", ".*:use-stock.*", Warning)
    warnings.filterwarnings(
        "ignore", r".*The property GtkAlignment:[^\s]+ is deprecated.*", Warning
    )

    settings = Gtk.Settings.get_default()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        settings.set_property("gtk-button-images", True)
        settings.set_property("gtk-menu-images", True)
    if hasattr(settings.props, "gtk_primary_button_warps_slider"):
        # https://bugzilla.gnome.org/show_bug.cgi?id=737843
        settings.set_property("gtk-primary-button-warps-slider", True)

    # Make sure PyGObject includes support for foreign cairo structs
    try:
        gi.require_foreign("cairo")
    except ImportError:
        print_e("PyGObject is missing cairo support")
        sys.exit(1)

    css_override = ThemeOverrider()

    if sys.platform == "darwin":
        # fix duplicated shadows for popups with Gtk+3.14
        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(b"""
            GtkWindow {
                box-shadow: none;
            }
            .tooltip {
                border-radius: 0;
                padding: 0;
            }
            .tooltip.background {
                background-clip: border-box;
            }
            """)
        css_override.register_provider("", style_provider)

    if gtk_version[:2] >= (3, 20):
        # https://bugzilla.gnome.org/show_bug.cgi?id=761435
        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(b"""
            spinbutton, button {
                min-height: 22px;
            }

            .view button {
                min-height: 24px;
            }

            entry {
                min-height: 28px;
            }

            entry.cell {
                min-height: 0;
            }
        """)
        css_override.register_provider("Adwaita", style_provider)
        css_override.register_provider("HighContrast", style_provider)

        # https://github.com/quodlibet/quodlibet/issues/2541
        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(b"""
            treeview.view.separator {
                min-height: 2px;
                color: @borders;
            }
        """)
        css_override.register_provider("Ambiance", style_provider)
        css_override.register_provider("Radiance", style_provider)
        # https://github.com/quodlibet/quodlibet/issues/2677
        css_override.register_provider("Clearlooks-Phenix", style_provider)
        # https://github.com/quodlibet/quodlibet/issues/2997
        css_override.register_provider("Breeze", style_provider)

    if gtk_version[:2] >= (3, 18):
        # Hack to get some grab handle like thing for panes
        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(b"""
            GtkPaned.vertical, paned.vertical >separator {
                -gtk-icon-source: -gtk-icontheme("view-more-symbolic");
                -gtk-icon-transform: rotate(90deg) scaleX(0.1) scaleY(3);
            }

            GtkPaned.horizontal, paned.horizontal >separator {
                -gtk-icon-source: -gtk-icontheme("view-more-symbolic");
                -gtk-icon-transform: rotate(0deg) scaleX(0.1) scaleY(3);
            }
        """)
        css_override.register_provider("", style_provider)

    # https://bugzilla.gnome.org/show_bug.cgi?id=708676
    warnings.filterwarnings("ignore", ".*g_value_get_int.*", Warning)

    # blacklist some modules, simply loading can cause segfaults
    sys.modules["gtk"] = None
    sys.modules["gpod"] = None
    sys.modules["gnome"] = None

    from quodlibet.qltk import pygobject_version, gtk_version

    MinVersions.GTK.check(gtk_version)
    MinVersions.PYGOBJECT.check(pygobject_version)


def _init_gst():
    """Call once before importing GStreamer"""

    arch_key = "64" if sys.maxsize > 2**32 else "32"
    registry_name = f"gst-registry-{sys.platform}-{arch_key}.bin"
    os.environ["GST_REGISTRY"] = os.path.join(get_cache_dir(), registry_name)

    assert "gi.repository.Gst" not in sys.modules

    import gi

    # We don't want python-gst, it changes API..
    assert "gi.overrides.Gst" not in sys.modules
    sys.modules["gi.overrides.Gst"] = None

    # blacklist some modules, simply loading can cause segfaults
    sys.modules["gst"] = None

    # We don't depend on Gst overrides, so make sure it's initialized.
    try:
        gi.require_version("Gst", "1.0")
        from gi.repository import Gst
    except (ValueError, ImportError):
        return

    if Gst.is_initialized():
        return

    from gi.repository import GLib

    try:
        ok, sys.argv[:] = Gst.init_check(sys.argv)
    except GLib.GError:
        print_e("Failed to initialize GStreamer")
        # Uninited Gst segfaults: make sure no one can use it
        sys.modules["gi.repository.Gst"] = None
    else:
        # monkey patching ahead
        _fix_gst_leaks()



# Copyright 2013 Simonas Kazlauskas
#        2016-22 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import json
from typing import Any
from collections.abc import Callable

from gi.repository import Soup, Gio, GLib, GObject
from gi.repository.GObject import ParamFlags, SignalFlags

from quodlibet.const import VERSION, WEBSITE
from quodlibet.util import print_d, print_w

PARAM_READWRITECONSTRUCT = (
    ParamFlags.CONSTRUCT_ONLY | ParamFlags.READABLE | ParamFlags.WRITABLE
)


class HTTPRequest(GObject.Object):
    """
    Class encapsulating a single HTTP request. These are meant to be sent
    and received only once. Behaviour is undefined otherwise.
    """

    __gsignals__ = {
        # Successes
        "sent": (SignalFlags.RUN_LAST, None, (Soup.Message,)),
        "received": (SignalFlags.RUN_LAST, None, (Gio.OutputStream,)),
        # Failures
        "send-failure": (SignalFlags.RUN_LAST, None, (object,)),
        "receive-failure": (SignalFlags.RUN_LAST, None, (object,)),
        # Common failure signal which will be emitted when either of above
        # failure signals are.
        "failure": (SignalFlags.RUN_LAST, None, (object,)),
    }

    message = GObject.Property(type=Soup.Message, flags=PARAM_READWRITECONSTRUCT)
    cancellable = GObject.Property(type=Gio.Cancellable, flags=PARAM_READWRITECONSTRUCT)
    istream = GObject.Property(type=Gio.InputStream, default=None)
    ostream = GObject.Property(type=Gio.OutputStream, default=None)

    def __init__(self, message, cancellable):
        if message is None:
            raise ValueError("Message may not be None")

        inner_cancellable = Gio.Cancellable()
        super().__init__(message=message, cancellable=inner_cancellable)
        if cancellable is not None:
            cancellable.connect(lambda *x: self.cancel(), None)

        self.connect("send-failure", lambda r, e: r.emit("failure", e))
        self.connect("receive-failure", lambda r, e: r.emit("failure", e))

        # For simple access
        self._receive_started = False
        self._uri = self.message.get_uri().to_string()

    def send(self):
        """
        Send the request and receive HTTP headers. Some of the body might
        get downloaded too.
        """
        session.send_async(self.message, 1, self.cancellable, self._sent, None)

    def _sent(self, session, task, data):
        m = self.message
        try:
            status = int(m.get_property("status-code"))
            if status >= 400:
                msg = f"HTTP {status} error in {m.get_method()} request to {self._uri}"
                print_w(msg)
                return self.emit("send-failure", Exception(msg))
            self.istream = session.send_finish(task)
            print_d(f"Got HTTP {status} on {m.get_method()} request to {self._uri}.")
            self.emit("sent", m)
        except GLib.GError as e:
            print_w(f"Failed sending {m.get_method()} request to {self._uri} ({e})")
            self.emit("send-failure", e)

    def provide_target(self, stream):
        if not stream:
            raise ValueError("Provided stream may not be None")
        if not self.ostream:
            self.ostream = stream
        else:
            raise RuntimeError("Only one output stream may be provided")

    def cancel(self):
        """
        Cancels the future and currently running HTTPRequest actions.

        It is safe to run this function before, during and after any action
        made with HTTPRequest.

        After HTTPRequest is cancelled, one usually would not do any more
        actions with it. However, it is safe to do something after
        cancellation, but those actions usually will fail.
        """

        if self.cancellable.is_cancelled():
            return False
        self.cancellable.cancel()

        # If we already have input stream, we can just close it, message
        # will come out as cancelled just fine.
        if self.istream and not self._receive_started:
            if not self.istream.is_closed():
                self.istream.close(None)
        return None

    def receive(self):
        """
        Receive data from the request into provided output stream. The request
        must be already sent, therefore this function will be usually called
        from the 'sent' signal handler.

        On completion of data receipt, HTTPRequest lifetime is ended and
        inner resources are cleaned up (except persistent connections that are
        part of session, not request).

        .. note::
        Be sure to clean up resources you've allocated yourself (e.g. close
        GOutputStreams, delete files on failure et cetera).
        """
        if not self.istream:
            raise RuntimeError("Cannot receive unsent request")
        if not self.ostream:
            raise RuntimeError("Cannot receive request without output stream")
        if self._receive_started:
            raise RuntimeError("Can receive only once")
        self._receive_started = True

        def spliced(ostream, task, data):
            try:
                ostream.splice_finish(task)
                self.istream.close(None)
                self.emit("received", ostream)
            except GLib.GError as e:
                self.istream.close(None)
                self.emit("receive-failure", e)

        # Do not ask splice to close the stream as Soup gets confused and
        # doesn't close connections
        # https://bugzilla.gnome.org/show_bug.cgi?id=711260
        flags = Gio.OutputStreamSpliceFlags.NONE
        self.ostream.splice_async(
            self.istream, flags, GLib.PRIORITY_DEFAULT, self.cancellable, spliced, None
        )


FailureCallback = Callable[[HTTPRequest, Exception, Any], None]


def download(
    message: Soup.Message,
    cancellable: Gio.Cancellable,
    callback: Callable,
    data: Any,
    try_decode: bool = False,
    failure_callback: FailureCallback | None = None,
):
    def received(request, ostream):
        ostream.close(None)
        bs = ostream.steal_as_bytes().get_data()
        if not try_decode:
            callback(message, bs, data)
            return
        # Otherwise try to decode data
        code = int(message.get_property("status-code"))
        if code >= 400:
            print_w("HTTP %d error received on %s" % (code, request._uri))
            return
        ctype = message.get_property("response-headers").get_content_type()
        encoding = ctype[1].get("charset", "utf-8")
        try:
            callback(message, bs.decode(encoding), data)
        except UnicodeDecodeError:
            callback(message, bs, data)

    request = HTTPRequest(message, cancellable)
    request.provide_target(Gio.MemoryOutputStream.new_resizable())
    request.connect("received", received)
    request.connect("sent", lambda r, m: r.receive())
    if failure_callback:
        request.connect("send-failure", failure_callback, data)
    request.send()


def download_json(
    message: Soup.Message,
    cancellable: Gio.Cancellable,
    callback: Callable,
    data: Any,
    failure_callback: FailureCallback | None = None,
):
    def cb(message, result, d):
        try:
            callback(message, json.loads(result), data)
        except ValueError:
            callback(message, None, data)

    download(message, cancellable, cb, None, True, failure_callback=failure_callback)


session = Soup.Session()
ua_string = f"Quodlibet/{VERSION} (+{WEBSITE})"
session.set_properties(user_agent=ua_string, timeout=15)



# Copyright 2019 LoveIsGrief
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
from quodlibet import plugins, qltk
from quodlibet.qltk.entry import UndoEntry
from quodlibet.util.thread import call_async_background, Cancellable

try:
    import musicbrainzngs
except ImportError as e:
    raise plugins.MissingModulePluginError("musicbrainzngs") from e

from gi.repository import Gtk
from quodlibet import _
from quodlibet.plugins import PluginConfig
from quodlibet.plugins.events import EventPlugin

ATTR_BRAINZ = "musicbrainz_trackid"
ATTR_RATING = "~#rating"

BRAINZ_APP = "quodlibetMusicBrainzSync"
VERSION = "0.1"

plugin_config = PluginConfig("musicbrainz-sync")
defaults = plugin_config.defaults
defaults.set("username", "")
defaults.set("password", "")


class MusicBrainzSyncPlugin(EventPlugin):
    PLUGIN_ID = "musicbrainzsync"
    VERSION = VERSION
    PLUGIN_NAME = _("MusicBrainz Sync")
    PLUGIN_DESC = _("Syncs the rating of a song with MusicBrainz.")

    def __init__(self):
        super().__init__()
        musicbrainzngs.set_rate_limit()
        musicbrainzngs.set_useragent(
            BRAINZ_APP, VERSION, "quodlibet@lists.sacredchao.net"
        )
        musicbrainzngs.auth(
            plugin_config.get("username"), plugin_config.get("password")
        )

    def plugin_on_changed(self, songs):
        ratings_dict = {
            song(ATTR_BRAINZ): int(song(ATTR_RATING) * 100)
            for song in songs
            if ATTR_BRAINZ in song and ATTR_RATING in song
        }
        if len(ratings_dict):
            call_async_background(
                musicbrainzngs.submit_ratings,
                Cancellable(),
                callback=lambda *args: (),
                kwargs={
                    "recording_ratings": ratings_dict,
                },
            )

    def PluginPreferences(self, parent):
        def changed(entry, key):
            if entry.get_property("sensitive"):
                plugin_config.set(key, entry.get_text())
                musicbrainzngs.auth(
                    plugin_config.get("username"), plugin_config.get("password")
                )

        box = Gtk.VBox(spacing=12)

        # first frame
        table = Gtk.Table(n_rows=5, n_columns=2)
        table.props.expand = False
        table.set_col_spacings(6)
        table.set_row_spacings(6)

        labels = []
        label_names = [_("User_name:"), _("_Password:")]
        for idx, name in enumerate(label_names):
            label = Gtk.Label(label=name)
            label.set_alignment(0.0, 0.5)
            label.set_use_underline(True)
            table.attach(
                label,
                0,
                1,
                idx,
                idx + 1,
                xoptions=Gtk.AttachOptions.FILL | Gtk.AttachOptions.SHRINK,
            )
            labels.append(label)

        row = 0

        # username
        entry = UndoEntry()
        entry.set_text(plugin_config.get("username"))
        entry.connect("changed", changed, "username")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        # password
        entry = UndoEntry()
        entry.set_text(plugin_config.get("password"))
        entry.set_visibility(False)
        entry.connect("changed", changed, "password")
        table.attach(entry, 1, 2, row, row + 1)
        labels[row].set_mnemonic_widget(entry)
        row += 1

        box.pack_start(qltk.Frame(_("Account"), child=table), True, True, 0)

        return box



# Copyright 2014 Christoph Reiter
#           2018 Ludovic Druette
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import config
from quodlibet.util import print_d

from ._base import MMKeysAction, MMKeysImportError


def iter_backends():
    if config.getboolean("settings", "disable_mmkeys"):
        return

    try:
        from .gnome import GnomeBackend, GnomeBackendOldName, MateBackend
    except MMKeysImportError:
        pass
    else:
        yield GnomeBackend
        yield GnomeBackendOldName
        yield MateBackend

    try:
        from .keybinder import KeybinderBackend
    except MMKeysImportError:
        pass
    else:
        yield KeybinderBackend

    try:
        from .winhook import WinHookBackend
    except MMKeysImportError:
        pass
    else:
        yield WinHookBackend

    try:
        from .osx import OSXBackend
    except MMKeysImportError:
        pass
    else:
        yield OSXBackend


def find_active_backend():
    print_d("Trying to find a mmkeys backend")
    for backend in iter_backends():
        if backend.is_active():
            print_d(f"Found {backend.__name__!r}")
            return backend
    return None


class MMKeysHandler:
    """Manages multiple keybinding backends and translates the generated
    events to actions on the player backend.
    """

    def __init__(self, app):
        self._backend = None
        self._window = app.window
        self._player = app.player
        self._player_options = app.player_options
        self._app_name = app.name

    def start(self):
        kind = find_active_backend()
        if not kind:
            return
        self._backend = kind(self._app_name, self._callback)
        # grab on start for cases when the window is hidden on start
        self._backend.grab()

        self._window.connect("notify::is-active", self._focus_event)

    def quit(self):
        if self._backend:
            self._backend.cancel()
            self._backend = None
            self._window = None
            self._player = None

    def _focus_event(self, window, param):
        if window.get_property(param.name) and self._backend:
            self._backend.grab()

    def _callback(self, action):
        print_d(f"Event {action!r} from {type(self._backend).__name__!r}")

        def seek_relative(seconds):
            current = player.get_position()
            current += seconds * 1000
            current = min(player.song("~#length") * 1000 - 1, current)
            current = max(0, current)
            player.seek(current)

        player = self._player
        player_options = self._player_options
        if action == MMKeysAction.PREV:
            player.previous(force=True)
        elif action == MMKeysAction.NEXT:
            player.next()
        elif action == MMKeysAction.STOP:
            player.stop()
        elif action == MMKeysAction.PLAY:
            player.play()
        elif action == MMKeysAction.PLAYPAUSE:
            player.playpause()
        elif action == MMKeysAction.PAUSE:
            player.paused = True
        elif action == MMKeysAction.FORWARD:
            if player.song:
                seek_relative(10)
        elif action == MMKeysAction.REWIND:
            if player.song:
                seek_relative(-10)
        elif action == MMKeysAction.REPEAT:
            player_options.repeat = not player_options.repeat
        elif action == MMKeysAction.SHUFFLE:
            player_options.shuffle = not player_options.shuffle
        else:
            assert 0, "unhandled event"



# Copyright 2020-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
import time

import quodlibet

DEVICE_CACHE_SEC = 600

try:
    from soco import SoCo, SoCoException
    from soco.data_structures import DidlMusicTrack, DidlPlaylistContainer, DidlItem
except ImportError as e:
    raise quodlibet.plugins.MissingModulePluginError("soco") from e

from collections.abc import Collection
from gi.repository import Gtk
from quodlibet import _
from quodlibet import app
from quodlibet import qltk
from quodlibet.formats import AudioFile
from quodlibet.plugins.playlist import PlaylistPlugin
from quodlibet.qltk import Icons, Dialog, WarningMessage
from quodlibet.qltk.notif import Task
from quodlibet.util import copool, escape
from quodlibet.util.dprint import print_d, print_w
from quodlibet.util.string.filter import remove_punctuation

try:
    import soco
except ImportError as e:
    raise quodlibet.plugins.MissingModulePluginError("soco") from e

PlaylistID = str
ID = str
Name = str
SonosPlaylistDict = dict[PlaylistID, Name]


class ComboBoxEntry(Gtk.ComboBox):
    def __init__(self, choices: dict[str, str], tooltip_markup=None):
        super().__init__(
            model=Gtk.ListStore(str, str), entry_text_column=1, has_entry=True
        )
        self._fill_model(choices)
        if tooltip_markup:
            self.get_child().set_tooltip_markup(tooltip_markup)

    def _fill_model(self, choices: dict[str, str]):
        self.clear()
        render = Gtk.CellRendererText()
        self.pack_start(render, True)
        self.add_attribute(render, "text", 1)

        model = self.get_model()
        for id_, name in choices.items():
            model.append(row=[id_, name])
        self.set_model(model)

        comp = Gtk.EntryCompletion()
        comp.set_model(self.get_model())
        comp.set_text_column(1)
        self.get_child().set_completion(comp)

    def get_chosen(self) -> tuple[ID | None, Name]:
        tree_iter = self.get_active_iter()
        if tree_iter is not None:
            model = self.get_model()
            id_, name = model[tree_iter][:2]
            return id_, name
        entry = self.get_child()
        return None, entry.get_text()

    def set_text(self, text: str):
        model = self.get_model()
        for i, (id_, value) in enumerate(model):
            if value == text:
                self.set_active(i)
                print_d(f"Text matched existing playlist: {id_} ({value!r})")
                return None

        return self.get_child().set_text(text)


class GetSonosPlaylistDialog(Dialog):
    def __init__(self, choices: SonosPlaylistDict):
        super().__init__(title="Which Sonos Playlist?", transient_for=None)
        self.options = choices

        self.set_border_width(6)
        self.set_resizable(True)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_OK"), Gtk.ResponseType.OK)
        self.vbox.set_spacing(6)
        self.set_default_response(Gtk.ResponseType.OK)

        box = Gtk.VBox(spacing=6)
        label = Gtk.Label(
            label=_(
                "Type a new playlist name,\n"
                "or choose an existing Sonos playlist to overwrite"
            )
        )
        box.set_border_width(6)
        label.set_line_wrap(True)
        label.set_justify(Gtk.Justification.CENTER)
        box.pack_start(label, True, True, 0)

        self._combo = ComboBoxEntry(choices)
        box.pack_start(self._combo, False, False, 0)

        self.vbox.pack_start(box, True, True, 0)
        self.get_child().show_all()

    def run(self, text: str | None = None) -> tuple[Name | None, str] | None:
        self.show()
        if text:
            self._combo.set_text(text)
        self._combo.grab_focus()
        resp = super().run()
        try:
            return self._combo.get_chosen() if resp == Gtk.ResponseType.OK else None
        finally:
            self.destroy()


class SonosPlaylistPlugin(PlaylistPlugin):
    PLUGIN_ID = "Export to Sonos Playlist"
    PLUGIN_NAME = _("Export to Sonos Playlist")
    PLUGIN_DESC = _("Exports a playlist to Sonos by matching tracks.")
    PLUGIN_ICON = Icons.NETWORK_WORKGROUP
    REQUIRES_ACTION = True

    DEBUG = False

    MAX_RESULTS_PER_SEARCH = 10
    """Request this many max results, for each search to then filter in code"""

    def __init__(self, playlists=None, library=None):
        super().__init__(playlists, library)
        self.__cancel = False
        self.devices = []
        self._last_discovery = 0

    def enabled(self):
        print_d(f"Using Sonos device: {self.device}")

    @property
    def device(self) -> SoCo | None:
        now = time.time()
        if not self.devices or now - self._last_discovery > DEVICE_CACHE_SEC:
            devices = soco.discover()
            print_d(f"Found {len(devices)} Sonos devices")
            coords = [d for d in devices if d.is_coordinator]
            if not coords:
                raise Exception(f"No Sonos coordinators found in {devices}")
            print_d(f"Found {len(coords)} Sonos coordinators")
            self.devices = coords
            self._last_discovery = now
        return self.devices[0]

    def __cancel_add(self):
        """Tell the copool to stop (adding songs)"""
        self.__cancel = True

    @classmethod
    def _score(cls, t: DidlMusicTrack, song: AudioFile) -> float:
        d = t.to_dict()
        try:
            person = song.list("~people")[0]
        except IndexError:
            person = None
        album = song("album")
        score = (
            int(
                remove_punctuation(t.title).lower()
                == remove_punctuation(song("title")).lower()
            )
            + int(bool(person) and person in d.values())
            + int(bool(album) and album in d.get("album", ""))
        )
        if cls.DEBUG:
            print_d(f"{score:.1f} for {t.title} ({d})")
        return score

    def __add_songs(
        self, task: Task, songs: Collection[AudioFile], spl: DidlPlaylistContainer
    ):
        """Generator for copool to add songs to the temp playlist"""
        assert self.device
        task_total = float(len(songs))
        print_d(
            "Adding %d song(s) to Sonos playlist. "
            "This might take a while..." % task_total
        )
        failures = []
        for i, song in enumerate(songs):
            if self.__cancel:
                print_d("Cancelled Sonos export")
                return
            lib = self.device.music_library

            # Exact title gets the best results it seems; some problems with ’
            search_term = song("title")
            assert search_term
            results = lib.get_tracks(
                search_term=search_term, max_items=self.MAX_RESULTS_PER_SEARCH
            )
            yield
            total = len(results)
            if total == 1:
                track = results[0]
            elif total > 1:
                desc = song("~~people~album~title")
                candidates = [(self._score(t, song), t) for t in results]
                in_order = sorted(candidates, reverse=True, key=lambda x: x[0])
                track = in_order[0][1]
                print_d(
                    f"From {len(results)} choice(s) for {desc!r}, "
                    f"chose {self.uri_for(track)}"
                )
            else:
                print_w(f'No results for "{search_term}"')
                failures.append(search_term)
                track = None
            if track:
                try:
                    self.device.add_item_to_sonos_playlist(track, spl)
                except SoCoException as e:
                    failures.append(track)
                    print_w(f"Couldn't add {track} ({e}, skipping")
            task.update(float(i) / task_total)
            yield
        task.update((task_total - 2) / task_total)
        yield
        task.finish()
        if failures:
            print_w(f"Got {len(failures)} failure(s), of {int(task_total)}: {failures}")
        successes = int(task_total) - len(failures)
        print_d(f"Finished export to {spl.title!r}, {successes} tracks in total")

    @staticmethod
    def uri_for(track: DidlItem) -> str:
        """More usable elsewhere (on Linux at least)"""
        return track.get_uri().replace("x-file-cifs", "smb")

    def plugin_playlist(self, playlist):
        print_d(f"Using Sonos device: {self.device!r}")
        device = self.device
        if not device:
            qltk.ErrorMessage(
                app.window,
                _("Error finding Sonos device(s)"),
                _("Error finding Sonos. Please check settings"),
            ).run()
        else:
            sonos_pls = device.get_sonos_playlists()
            pl_id_to_name = {pl.item_id: pl.title for pl in sonos_pls}
            print_d(f"Sonos playlists: {pl_id_to_name}")
            ret = GetSonosPlaylistDialog(pl_id_to_name).run(playlist.name)
            if ret:
                spl_id, name = ret
                if spl_id:
                    spl: DidlPlaylistContainer = next(
                        s for s in sonos_pls if s.item_id == spl_id
                    )
                    print_w(f"Replacing existing Sonos playlist {spl_id} ({spl.title})")
                    try:
                        device.remove_sonos_playlist(spl)
                        print_d("Removed existing playlist OK")
                    except SoCoException as e:
                        tmpl = _("Failed to delete existing Sonos playlist %s:")
                        err_str = f" \n<tt>{escape(e)!r}</tt>"
                        dialog = WarningMessage(
                            None, tmpl % escape(spl.title) + err_str, escape_desc=False
                        )
                        dialog.run()
                        return

                print_d(f"Creating new playlist {name!r}")
                spl = device.create_sonos_playlist(name)
                data = {"playlist": name[:30], "total": len(playlist)}
                task = Task(
                    "Sonos",
                    _("Export to playlist %(playlist)r (%(total)d tracks)") % data,
                    stop=self.__cancel_add,
                )
                copool.add(
                    self.__add_songs,
                    task,
                    playlist.songs,
                    spl,
                    funcid="sonos-playlist-save",
                )



# Copyright 2005 Joe Wreschnig
#           2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _
from quodlibet.qltk import Icons
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.plugins.songshelpers import any_song, is_writable


class ForceWrite(SongsMenuPlugin):
    PLUGIN_ID = "Force Write"
    PLUGIN_NAME = _("Update Tags in Files")
    PLUGIN_DESC = _(
        "Update modified tags in files. "
        "This will ensure play counts and ratings are up to date."
    )
    PLUGIN_ICON = Icons.DOCUMENT_SAVE

    plugin_handles = any_song(is_writable)

    def plugin_song(self, song):
        song._needs_write = True



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ._pattern import (
    Pattern,
    FileFromPattern,
    XMLFromPattern,
    XMLFromMarkupPattern,
    Error,
    ArbitraryExtensionFileFromPattern,
    URLFromPattern,
)


URLFromPattern
ArbitraryExtensionFileFromPattern
FileFromPattern
Pattern
XMLFromMarkupPattern
XMLFromPattern
Error



# Copyright (C) 2006 - Steve Frécinaux
#            2016-23 - Nick Boultbee
#               2021 - halfbrained@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Parts from "Interactive Python-GTK Console"
# (stolen from epiphany's console.py)
#     Copyright (C), 1998 James Henstridge <james@daa.com.au>
#     Copyright (C), 2005 Adam Hooper <adamh@densi.com>
# Bits from gedit Python Console Plugin
#     Copyright (C), 2005 Raphaël Slinckx

# PythonConsole taken from totem
# Plugin parts:
# Copyright 2009,2010,2013 Christoph Reiter
#                     2016 Nick Boultbee

import sys
import re
import traceback
from itertools import takewhile

from gi.repository import Gtk, Pango, Gdk, GLib

from quodlibet import _, app, ngettext
from quodlibet import const
from quodlibet.plugins.events import EventPlugin
from quodlibet.plugins.gui import UserInterfacePlugin
from quodlibet.qltk import Icons, add_css, Align
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.util.collection import Collection
from quodlibet.util import print_


class PyConsole(SongsMenuPlugin):
    PLUGIN_ID = "Python Console"
    PLUGIN_NAME = _("Python Console")
    PLUGIN_DESC = _("Interactive Python console. Opens a new window.")
    PLUGIN_ICON = Icons.UTILITIES_TERMINAL

    def plugin_songs(self, songs):
        desc = ngettext("%d song", "%d songs", len(songs)) % len(songs)
        win = ConsoleWindow(create_console(songs), title=desc)
        win.set_icon_name(self.PLUGIN_ICON)
        win.set_title(
            _("{plugin_name} for {songs} ({app})").format(
                plugin_name=self.PLUGIN_NAME, songs=desc, app=app.name
            )
        )
        win.show_all()


class PyConsoleSidebar(EventPlugin, UserInterfacePlugin):
    PLUGIN_ID = "Python Console Sidebar"
    PLUGIN_NAME = _("Python Console Sidebar")
    PLUGIN_DESC = _(
        "Interactive Python console sidebar, "
        "that follows the selected songs in the main window."
    )
    PLUGIN_ICON = Icons.UTILITIES_TERMINAL

    def enabled(self):
        self.console = create_console()

    def plugin_on_songs_selected(self, songs):
        self.console.namespace = namespace_for(songs)

    def create_sidebar(self):
        align = Align(self.console)
        self.sidebar = align
        self.sidebar.show_all()
        return align


def create_console(songs=None):
    console = PythonConsole(namespace_for(songs)) if songs else PythonConsole()
    access_string = _("You can access the following objects by default:")
    access_string += (
        "\\n"
        "%7s: SongWrapper objects\\n"
        "%7s: Song dictionaries\\n"
        "%7s: Filename list\\n"
        "%7s: Songs Collection\\n"
        "%7s: Application instance" % ("songs", "sdict", "files", "col", "app")
    )
    dir_string = _("Your current working directory is:")

    console.eval("import mutagen", False)
    console.eval("import os", False)
    python_major_ver = sys.version.split()[0]
    console.eval(
        f'print("Python: {python_major_ver} / Quod Libet: {const.VERSION}")', False
    )
    console.eval(f'print("{access_string}")', False)
    console.eval(f'print("{dir_string} "+ os.getcwd())', False)
    return console


def namespace_for(song_wrappers):
    files = [song("~filename") for song in song_wrappers]
    song_dicts = [song._song for song in song_wrappers]
    collection = Collection()
    collection.songs = song_dicts
    return {
        "songs": song_wrappers,
        "files": files,
        "sdict": song_dicts,
        "col": collection,
        "app": app,
    }


class ConsoleWindow(Gtk.Window):
    def __init__(self, console, title=None):
        Gtk.Window.__init__(self)
        if title:
            self.set_title(title)
        self.add(console)
        self.set_size_request(700, 500)
        console.connect("destroy", lambda *x: self.destroy())


class PythonConsole(Gtk.ScrolledWindow):
    def __init__(self, namespace=None, destroy_cb=None):
        Gtk.ScrolledWindow.__init__(self)

        self.destroy_cb = destroy_cb
        self.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.set_shadow_type(Gtk.ShadowType.NONE)
        self.view = Gtk.TextView()
        add_css(
            self,
            "scrolledwindow { padding: 6px; "
            "background-color: white; background-color: @content_view_bg;}",
        )
        self.view.modify_font(Pango.font_description_from_string("Monospace"))
        self.view.set_editable(True)
        self.view.set_wrap_mode(Gtk.WrapMode.CHAR)
        self.add(self.view)
        self.view.show()

        buffer = self.view.get_buffer()
        self.normal = buffer.create_tag("normal")
        self.error = buffer.create_tag("error")
        self.error.set_property("foreground", "red")
        self.command = buffer.create_tag("command")
        self.command.set_property("foreground", "blue")

        self.__spaces_pattern = re.compile(r"^\s+")
        self.namespace = namespace or {}

        self.block_command = False

        # Init first line
        buffer.create_mark("input-line", buffer.get_end_iter(), True)
        buffer.insert(buffer.get_end_iter(), ">>> ")
        buffer.create_mark("input", buffer.get_end_iter(), True)

        # Init history
        self.history = [""]
        self.history_pos = 0
        self.current_command = ""
        self.namespace["__history__"] = self.history

        # Set up hooks for standard output.
        self.stdout = OutFile(self, self.normal)
        self.stderr = OutFile(self, self.error)

        # Signals
        self.view.connect("key-press-event", self.__key_press_event_cb)
        buffer.connect("mark-set", self.__mark_set_cb)

    def __key_press_event_cb(self, view, event):
        modifier_mask = Gtk.accelerator_get_default_mod_mask()
        event_state = event.state & modifier_mask

        if event.keyval == Gdk.KEY_d and event_state == Gdk.ModifierType.CONTROL_MASK:
            self.destroy()

        elif event.keyval == Gdk.KEY_Return and (
            event_state & (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        ):
            # Get the command
            buffer = view.get_buffer()
            inp_mark = buffer.get_mark("input")
            inp = buffer.get_iter_at_mark(inp_mark)
            cur = buffer.get_end_iter()
            line = buffer.get_text(inp, cur, True)
            self.current_command = self.current_command + line + "\n"
            self.history_add(line)

            # Prepare the new line
            cur = buffer.get_end_iter()
            buffer.insert(cur, "\n... ")
            cur = buffer.get_end_iter()
            buffer.move_mark(inp_mark, cur)

            # Keep indentation of preceding line
            spaces = re.match(self.__spaces_pattern, line)
            if spaces is not None:
                buffer.insert(cur, line[spaces.start() : spaces.end()])
            else:
                buffer.insert(cur, "    ")
            cur = buffer.get_end_iter()

            buffer.place_cursor(cur)
            GLib.idle_add(self.scroll_to_end)
            return True

        elif event.keyval == Gdk.KEY_Return:
            # Get the marks
            buffer = view.get_buffer()
            lin_mark = buffer.get_mark("input-line")
            inp_mark = buffer.get_mark("input")

            # Get the command line
            inp = buffer.get_iter_at_mark(inp_mark)
            cur = buffer.get_end_iter()
            line = buffer.get_text(inp, cur, True)
            self.current_command = self.current_command + line + "\n"

            # Make the line blue
            lin = buffer.get_iter_at_mark(lin_mark)
            buffer.apply_tag(self.command, lin, cur)
            buffer.insert(cur, "\n")

            cur_strip = self.current_command.rstrip()

            if cur_strip.endswith(":") or (
                self.current_command[-2:].strip() != "" and self.block_command
            ):
                # Unfinished block command
                self.block_command = True
                com_mark = "... "
            elif cur_strip.endswith("\\"):
                com_mark = "... "
            else:
                # Eval the command
                self.history_add(cur_strip)
                self.__run(self.current_command)
                self.current_command = ""
                self.block_command = False
                com_mark = ">>> "

            # Prepare the new line
            cur = buffer.get_end_iter()
            buffer.move_mark(lin_mark, cur)
            buffer.insert(cur, com_mark)
            cur = buffer.get_end_iter()
            buffer.move_mark(inp_mark, cur)

            # Keep indentation of preceding line
            if com_mark == "... ":
                spaces = re.match(self.__spaces_pattern, line)
                if spaces is not None:
                    buffer.insert(cur, line[spaces.start() : spaces.end()])
                if cur_strip.endswith(":"):
                    buffer.insert(cur, "    ")

            buffer.place_cursor(cur)
            GLib.idle_add(self.scroll_to_end)
            return True

        elif event.keyval == Gdk.KEY_c and event_state == Gdk.ModifierType.CONTROL_MASK:
            # Get the marks
            buffer = view.get_buffer()
            lin_mark = buffer.get_mark("input-line")
            inp_mark = buffer.get_mark("input")
            cur = buffer.get_end_iter()

            # New line
            buffer.insert(cur, "\n")
            com_mark = ">>> "
            cur = buffer.get_end_iter()
            buffer.move_mark(lin_mark, cur)
            buffer.insert(cur, com_mark)

            # Move marks
            cur = buffer.get_end_iter()
            buffer.move_mark(inp_mark, cur)
            buffer.place_cursor(cur)
            GLib.idle_add(self.scroll_to_end)
            return True

        elif event.keyval == Gdk.KEY_KP_Down or event.keyval == Gdk.KEY_Down:
            # Next entry from history
            view.emit_stop_by_name("key_press_event")
            self.history_down()
            GLib.idle_add(self.scroll_to_end)
            return True

        elif event.keyval == Gdk.KEY_KP_Up or event.keyval == Gdk.KEY_Up:
            # Previous entry from history
            view.emit_stop_by_name("key_press_event")
            self.history_up()
            GLib.idle_add(self.scroll_to_end)
            return True

        elif (
            event.keyval == Gdk.KEY_KP_Left
            or event.keyval == Gdk.KEY_Left
            or event.keyval == Gdk.KEY_BackSpace
        ):
            buffer = view.get_buffer()
            inp = buffer.get_iter_at_mark(buffer.get_mark("input"))
            cur = buffer.get_iter_at_mark(buffer.get_insert())
            return inp.compare(cur) == 0

        elif event.keyval == Gdk.KEY_Home:
            # Go to the begin of the command instead of the begin of the line
            buffer = view.get_buffer()
            inp = buffer.get_iter_at_mark(buffer.get_mark("input"))
            if event_state == Gdk.ModifierType.SHIFT_MASK:
                buffer.move_mark_by_name("insert", inp)
            else:
                buffer.place_cursor(inp)
            return True

        # completion - Tab, Shift+Tab, Ctrl+Space , Ctrl+Shift+Space
        elif (event.keyval == Gdk.KEY_Tab or event.keyval == Gdk.KEY_ISO_Left_Tab) or (
            event.keyval == Gdk.KEY_space
            and (
                event_state
                == (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
                or event_state == (Gdk.ModifierType.CONTROL_MASK)
            )
        ):
            buffer = view.get_buffer()

            # get string to left of caret  inside command text-line
            _inp_mark = buffer.get_mark("input")
            _ins_mark = buffer.get_mark("insert")
            _inp = buffer.get_iter_at_mark(_inp_mark)
            ins = buffer.get_iter_at_mark(_ins_mark)
            cmd_start = buffer.get_text(_inp, ins, True)

            # get identifiers chain string, e.g. `a.b.c`, or a single identifier
            _identifiers_chars = takewhile(
                lambda x: x.isalnum() or x in {"_", "."}, reversed(cmd_start)
            )
            _idcs_len = len(
                list(_identifiers_chars)
            )  # lengt of identifiers chain string
            ids_str = cmd_start[-_idcs_len:]

            # Ctrl+Shift+Space: has Shift: include identifiers starting with '__'
            is_shift = bool(event_state & Gdk.ModifierType.SHIFT_MASK)

            comp_items = self.get_completion_items(ids_str, include_private=is_shift)

            choice = None
            if len(comp_items) > 1:
                # sort completions: case-insensitive,
                # items starting with '_' - at the end
                comp_items.sort(key=lambda x: (x[0].startswith("_"), x[0].lower()))

                dialog = ListChoiceDialog(self.get_parent(), comp_items)
                choice = dialog.run()
                dialog.destroy()
            elif len(comp_items) == 1:
                # if current text is only suggestion, add a '.' if object
                # has properties visible to autosuggestion
                last = ids_str.split(".")[-1]
                if last == comp_items[0][0]:
                    next_comp_items = self.get_completion_items(
                        ids_str + ".", include_private=is_shift
                    )
                    if next_comp_items:
                        buffer.insert(ins, ".")
                else:
                    choice = 0
            elif "\n" in self.current_command:
                # if we're doing multi-line editing, tab should add 4 spaces
                buffer.insert(ins, "    ")

            if isinstance(choice, int) and choice >= 0:
                last = ids_str.split(".")[-1]
                insert_text = comp_items[choice][0]
                if last:  # cut off prefix, already present
                    insert_text = insert_text[len(last) :]

                buffer.insert(ins, insert_text)

            return True
        return None

    def __mark_set_cb(self, buffer, iter, name):
        input = buffer.get_iter_at_mark(buffer.get_mark("input"))
        pos = buffer.get_iter_at_mark(buffer.get_insert())
        self.view.set_editable(pos.compare(input) != -1)

    def get_command_line(self):
        buffer = self.view.get_buffer()
        inp = buffer.get_iter_at_mark(buffer.get_mark("input"))
        cur = buffer.get_end_iter()
        return buffer.get_text(inp, cur, True)

    def set_command_line(self, command):
        buffer = self.view.get_buffer()
        mark = buffer.get_mark("input")
        inp = buffer.get_iter_at_mark(mark)
        cur = buffer.get_end_iter()
        buffer.delete(inp, cur)
        buffer.insert(inp, command)
        buffer.select_range(buffer.get_iter_at_mark(mark), buffer.get_end_iter())
        self.view.grab_focus()

    def history_add(self, line):
        if line.strip() != "":
            self.history_pos = len(self.history)
            self.history[self.history_pos - 1] = line
            self.history.append("")

    def history_up(self):
        if self.history_pos > 0:
            self.history[self.history_pos] = self.get_command_line()
            self.history_pos -= 1
            self.set_command_line(self.history[self.history_pos])

    def history_down(self):
        if self.history_pos < len(self.history) - 1:
            self.history[self.history_pos] = self.get_command_line()
            self.history_pos += 1
            self.set_command_line(self.history[self.history_pos])

    def scroll_to_end(self):
        iter = self.view.get_buffer().get_end_iter()
        self.view.scroll_to_iter(iter, 0.0, False, 0.5, 0.5)
        return False

    def write(self, text, tag=None):
        buf = self.view.get_buffer()
        if tag is None:
            buf.insert(buf.get_end_iter(), text)
        else:
            buf.insert_with_tags(buf.get_end_iter(), text, tag)

        GLib.idle_add(self.scroll_to_end)

    def eval(self, command, display_command=False):
        buffer = self.view.get_buffer()
        lin = buffer.get_mark("input-line")
        buffer.delete(buffer.get_iter_at_mark(lin), buffer.get_end_iter())

        if isinstance(command, list | tuple):
            for c in command:
                if display_command:
                    self.write(">>> " + c + "\n", self.command)
                self.__run(c)
        else:
            if display_command:
                self.write(">>> " + c + "\n", self.command)
            self.__run(command)

        cur = buffer.get_end_iter()
        buffer.move_mark_by_name("input-line", cur)
        buffer.insert(cur, ">>> ")
        cur = buffer.get_end_iter()
        buffer.move_mark_by_name("input", cur)
        self.view.scroll_to_iter(buffer.get_end_iter(), 0.0, False, 0.5, 0.5)

    def __run(self, command):
        sys.stdout, self.stdout = self.stdout, sys.stdout
        sys.stderr, self.stderr = self.stderr, sys.stderr

        try:
            try:
                r = eval(command, self.namespace, self.namespace)
                if r is not None:
                    print_(repr(r))
            except SyntaxError:
                exec(command, self.namespace)
        except Exception:
            if hasattr(sys, "last_type") and sys.last_type is SystemExit:
                self.destroy()
            else:
                traceback.print_exc()

        sys.stdout, self.stdout = self.stdout, sys.stdout
        sys.stderr, self.stderr = self.stderr, sys.stderr

    def get_completion_items(self, ids_str, include_private=False):
        """get completion suggestions
        ids_str - identifiers chain string, e.g. `a.b.c`, or a single identifier
        """
        import inspect

        def get_comp(obj, pre):
            """get completion item names from `obj`-object or `self.namespace`
            with prefix `pre`
            """
            dir_result = dir(obj) if obj is not None else self.namespace

            comp = []  # result: (completion-name, details)
            for name in dir_result:
                if (pre and not name.startswith(pre)) or (
                    not include_private and name.startswith("__")
                ):
                    continue

                if obj is not None:
                    try:
                        f = getattr(obj, name)
                    except Exception:
                        continue
                else:
                    f = self.namespace.get(name)

                if not callable(f):
                    comp.append((name, ""))
                else:
                    try:
                        spec = inspect.getfullargspec(f)
                    except TypeError:
                        spec = None

                    if spec:
                        sargs = []
                        arglen = len(spec.args) if spec.args else 0
                        deflen = len(spec.defaults) if spec.defaults else 0
                        noargs = arglen - deflen
                        for i in range(arglen):
                            if i == 0 and spec.args[i] == "self":
                                continue

                            if i < noargs:
                                sargs.append(spec.args[i])
                            else:
                                default_arg = repr(spec.defaults[i - noargs])
                                sargs.append(spec.args[i] + "=" + default_arg)

                        details = " ({})".format(", ".join(sargs))
                        comp.append((name, details))
                    else:
                        comp.append((name, " ()"))
            # end for

            return comp

        # end get_comp()

        comp = None
        # method, field
        if "." in ids_str:
            spl = ids_str.split(".")

            # search for target object
            var = None
            for fname in spl[:-1]:
                if var is None:
                    var = self.namespace.get(fname)
                else:
                    try:
                        var = getattr(var, fname, None)
                    except Exception:
                        pass

                if var is None:
                    break

            if var is not None:
                comp = get_comp(obj=var, pre=spl[-1])

        # single var or empty
        else:
            comp = get_comp(obj=None, pre=ids_str)

        return comp or []


class OutFile:
    """A fake output file object. It sends output to a TK test widget,
    and if asked for a file number, returns one set on instance creation"""

    def __init__(self, console, tag):
        self.console = console
        self.tag = tag

    def close(self):
        pass

    def flush(self):
        pass

    def fileno(self):
        raise OSError

    def isatty(self):
        return 0

    def read(self, a):
        return ""

    def readline(self):
        return ""

    def readlines(self):
        return []

    def write(self, s):
        self.console.write(s, self.tag)

    def writelines(self, l):
        self.console.write(l, self.tag)

    def seek(self, a):
        raise OSError(29, "Illegal seek")

    def tell(self):
        raise OSError(29, "Illegal seek")

    truncate = tell


class ListChoiceDialog(Gtk.Dialog):
    """display listbox to choose an item from `rows` list,
    returns index if the chosen item, if positive
    """

    def __init__(self, parent, rows):
        Gtk.Dialog.__init__(self, title=_("Completion"), transient_for=parent, flags=0)
        self.set_default_size(400, 250)

        listbox = Gtk.ListBox()
        listbox.set_selection_mode(Gtk.SelectionMode.SINGLE)

        for i, (name, details) in enumerate(rows):
            row = Gtk.ListBoxRow()
            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            row.add(hbox)

            lbl = Gtk.Label(label=name, xalign=0)
            lbl2 = Gtk.Label(label=details, xalign=0)
            hbox.pack_start(lbl, expand=False, fill=False, padding=0)
            hbox.pack_start(lbl2, expand=True, fill=True, padding=0)

            # dim the details-label
            style = lbl2.get_style_context()
            style.add_class("dim-label")
            add_css(lbl2, ".dim-label { opacity: 0.5; } ")

            listbox.add(row)

            if i == 0:
                listbox.set_focus_child(row)

        listbox.connect("row-activated", self.on_row_click)

        scroll = Gtk.ScrolledWindow()
        scroll.add(listbox)

        content = self.get_content_area()
        content.pack_start(scroll, True, True, 0)

        content.show_all()
        self.get_action_area().hide()

    def on_row_click(self, listbox, row):
        self.response(row.get_index())



# Copyright 2016 Mice Pápai
#           2018 Nick Boultbee
#
# Based on lastfm.py by Simonas Kazlauskas
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from os import path

import re
from gi.repository import Soup

from quodlibet import _
from quodlibet.plugins.cover import cover_dir
from quodlibet.util.http import download_json
from quodlibet.util.cover.http import escape_query_value
from quodlibet.util.cover.http import ApiCoverSourcePlugin
from quodlibet.util.path import escape_filename
from quodlibet.util import print_d


class DiscogsCover(ApiCoverSourcePlugin):
    PLUGIN_ID = "discogs-cover"
    PLUGIN_NAME = _("Discogs Cover Source")
    PLUGIN_DESC = _("Downloads covers from Discogs.")

    credentials = (
        "key=aWfZGjHQvkMcreUECGAp" + "&secret=VlORkklpdvAwJMwxUjNNSgqicjuizJAl"
    )
    use_secondary = True

    @classmethod
    def group_by(cls, song):
        return song.album_key

    @staticmethod
    def priority():
        return 0.6

    @property
    def cover_path(self):
        mbid = self.song.get("musicbrainz_albumid", None)
        if mbid:
            return path.join(cover_dir, escape_filename(mbid))
        return super().cover_path

    @property
    def url(self):
        _url = (
            "https://api.discogs.com/database/search?"
            + self.credentials
            + "&format=CD&per_page=5"
            "&type=release" + "&artist={artist}" + "&release_title={album}"
        )
        # Discogs seems to use 'Various' almost exclusively for compilations
        artists = self._album_artists_for(self.song) or "Various"
        if "various artists" in artists.lower():
            artists = "Various"
        artist = escape_query_value(artists)
        album = escape_query_value(self.song.get("album", ""))
        if artist and album:
            return _url.format(artist=artist, album=album)
        return None  # Not enough data

    def _handle_search_response(self, message, json_dict, data=None):
        if not json_dict:
            print_d("Server did not return any valid JSON")
            return self.emit("search-complete", [])
        try:
            results = json_dict.get("results", [])
            covers = filter(
                None, (self.result_for(r.get("cover_image", None)) for r in results)
            )
            if covers:
                return self.emit("search-complete", list(covers))
            res_url = results[0].get("resource_url", "")
        except IndexError:
            return self.emit("search-complete", [])
        else:
            msg = Soup.Message.new("GET", f"{res_url}?{self.credentials}")
            download_json(msg, self.cancellable, self._handle_album_data, None)

    def _handle_album_data(self, message, json_dict, data=None):
        images = json_dict.get("images", None)

        if not images:
            print_d("Covers are not available")
            return self.emit("search-complete", [])

        results = list(self._covers_of_type(images))
        if not results and images and self.use_secondary:
            results.append(next(self._covers_of_type(images, "secondary")))

        self.emit("search-complete", results)
        return None

    def _covers_of_type(self, images, image_type="primary"):
        for image in images:
            url = image.get("uri")
            if url and image["type"] == image_type:
                yield self.result_for(url)

    def result_for(self, url):
        if not url:
            return None
        dimensions = re.compile(r"/(\d+x\d+)/").search(url)
        dimensions = dimensions and dimensions.group(1)
        if dimensions:
            dims = map(int, dimensions.split("x"))
            if min(dims) < self.MIN_DIMENSION:
                print_d(f"{dimensions} is too small to use")
                return None
        return {"cover": url, "dimensions": dimensions}



# Copyright 2004 Joe Wreschnig, Michael Urman
#           2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet.player._base import BasePlayer
from quodlibet.player import PlayerError


class NullPlayer(BasePlayer):
    version_info = "Null Player Backend"
    name = "Null"

    def __init__(self, sinkname="", librarian=None):
        super().__init__()
        self._paused = True
        self._source = None
        self._volume = 1.0
        self._mute = False
        self._position = 0

    def _destroy(self):
        pass

    def get_position(self):
        """Return the current playback position in milliseconds,
        or 0 if no song is playing."""
        return self._position

    @property
    def paused(self):
        return self._paused

    @paused.setter
    def paused(self, paused):
        if paused == self._paused:
            return
        self._paused = paused
        self.emit((paused and "paused") or "unpaused")

    def do_get_property(self, property):
        if property.name == "volume":
            return self._volume
        if property.name == "seekable":
            if self.song is None:
                return False
            return True
        if property.name == "mute":
            return self._mute
        raise AttributeError

    def do_set_property(self, property, v):
        if property.name == "volume":
            self._volume = v
        elif property.name == "mute":
            self._mute = v
        else:
            raise AttributeError

    def _error(self, message):
        self.paused = True
        self.emit("error", self.song, PlayerError(message))

    def seek(self, pos):
        """Seek to a position in the song, in milliseconds."""

        self._position = pos
        if self.song:
            self.emit("seek", self.song, pos)

    def _end(self, stopped, next_song=None):
        # We need to set self.song to None before calling our signal
        # handlers. Otherwise, if they try to end the song they're given
        # (e.g. by removing it), then we get in an infinite loop.
        song = self.song
        self.song = self.info = None
        self.emit("song-ended", song, stopped)

        current = next_song if next_song else (self._source and self._source.current)

        # Then, set up the next song.
        self._position = 0
        self.song = self.info = current
        self.emit("song-started", self.song)

        if self.song is None:
            self.paused = True

        if not self.paused and song is None:
            self.emit("unpaused")

        # seekable might change if we change to None, so notify just in case
        self.notify("seekable")

    def can_play_uri(self, uri):
        return True


def init(librarian):
    return NullPlayer(librarian)



# Copyright 2011,2013 Christoph Reiter
#             2016-22 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk, Pango, GLib

from quodlibet import _, util
from quodlibet.qltk import Button, Window
from quodlibet.util import connect_obj, print_w

from .acoustid import AcoustidSubmissionThread
from .analyze import FingerPrintPool


def get_stats(results):
    got_mbid = got_meta = 0

    for result in results:
        song = result.song
        got_mbid += bool(song("musicbrainz_trackid"))
        got_meta += bool(song("artist") and song.get("title") and song("album"))

    return got_mbid, got_meta


def can_submit(result):
    got_mbid, got_meta = get_stats([result])
    return bool(got_mbid or got_meta)


class FingerprintDialog(Window):
    def __init__(self, songs):
        super().__init__()
        self.set_border_width(12)
        self.set_title(_("Submit Acoustic Fingerprints"))
        self.set_default_size(450, 0)

        outer_box = Gtk.VBox(spacing=12)

        box = Gtk.VBox(spacing=6)

        self.__label = label = Gtk.Label()
        label.set_markup(util.bold(_("Generating fingerprints:")))
        label.set_alignment(0, 0.5)
        box.pack_start(label, False, True, 0)

        self.__bar = bar = Gtk.ProgressBar()
        self.__set_fraction(0)
        box.pack_start(bar, False, True, 0)
        self.__label_song = label_song = Gtk.Label()
        label_song.set_alignment(0, 0.5)
        label_song.set_ellipsize(Pango.EllipsizeMode.MIDDLE)
        box.pack_start(label_song, False, True, 0)

        self.__stats = stats = Gtk.Label()
        stats.set_alignment(0, 0.5)
        stats.set_line_wrap(True)
        stats.set_size_request(426, -1)
        expand = Gtk.Expander.new_with_mnemonic(_("_Details"))
        expand.set_resize_toplevel(True)
        expand.add(stats)

        def expand_cb(expand, *args):
            self.resize(self.get_size()[0], 1)

        stats.connect("unmap", expand_cb)

        box.pack_start(expand, False, False, 0)

        self.__fp_results = {}
        self.__fp_done = 0
        self.__songs = songs
        self.__musicdns_thread = None
        self.__acoustid_thread = None

        self.__update_stats()

        pool = FingerPrintPool()

        bbox = Gtk.HButtonBox()
        bbox.set_layout(Gtk.ButtonBoxStyle.END)
        bbox.set_spacing(6)
        self.__submit = submit = Button(_("_Submit"))
        submit.set_sensitive(False)
        submit.connect("clicked", self.__submit_cb)
        cancel = Button(_("_Cancel"))
        connect_obj(cancel, "clicked", self.__cancel_cb, pool)
        bbox.pack_start(cancel, True, True, 0)
        bbox.pack_start(submit, True, True, 0)

        outer_box.pack_start(box, True, True, 0)
        outer_box.pack_start(bbox, False, True, 0)

        pool.connect("fingerprint-done", self.__fp_done_cb)
        pool.connect("fingerprint-error", self.__fp_error_cb)
        pool.connect("fingerprint-started", self.__fp_started_cb)

        for song in songs:
            pool.push(song)

        connect_obj(self, "delete-event", self.__cancel_cb, pool)

        self.add(outer_box)
        self.show_all()

    def __update_stats(self):
        all_ = len(self.__songs)
        results = self.__fp_results.values()
        to_send = len(list(filter(can_submit, results)))
        valid_fp = len(results)
        got_mbid, got_meta = get_stats(results)

        text = _(
            "Songs either need a <i><b>musicbrainz_trackid</b></i>, "
            "or <i><b>artist</b></i> / "
            "<i><b>title</b></i> / <i><b>album</b></i> tags to get submitted."
        )
        text += "\n\n" + util.italic(_("Fingerprints:"))
        text += " %d/%d" % (valid_fp, all_)
        text += "\n" + util.italic(_("Songs with MBIDs:"))
        text += " %d/%d" % (got_mbid, all_)
        text += "\n" + util.italic(_("Songs with sufficient tags:"))
        text += " %d/%d" % (got_meta, all_)
        text += "\n" + util.italic(_("Songs to submit:"))
        text += " %d/%d" % (to_send, all_)
        self.__stats.set_markup(text)

    def __set_fraction(self, progress):
        self.__bar.set_fraction(progress)
        self.__bar.set_text("%d%%" % round(progress * 100))

    def __inc_fp_fraction(self):
        self.__fp_done += 1
        frac = self.__fp_done / float(len(self.__songs))
        self.__set_fraction(frac)
        if self.__fp_done == len(self.__songs):
            self.__submit.set_sensitive(True)
            self.__show_final_stats()

    def __fp_started_cb(self, pool, song):
        # increase by an amount smaller than one song, so that the user can
        # see some progress from the beginning.
        self.__set_fraction(0.5 / len(self.__songs) + self.__bar.get_fraction())
        self.__label_song.set_text(song("~filename"))

    def __fp_done_cb(self, pool, result):
        self.__fp_results[result.song] = result
        self.__inc_fp_fraction()
        self.__update_stats()

    def __fp_error_cb(self, pool, song, error):
        print_w("[fingerprint] " + error)
        self.__inc_fp_fraction()
        self.__update_stats()

    def __show_final_stats(self):
        all_ = len(self.__songs)
        results = self.__fp_results.values()
        to_send = len(list(filter(can_submit, results)))
        self.__label_song.set_text(
            _("Done. %(to-send)d/%(all)d songs to submit.")
            % {"to-send": to_send, "all": all_}
        )

    def __cancel_cb(self, pool, *args):
        self.destroy()

        def idle_cancel():
            pool.stop()
            if self.__acoustid_thread:
                self.__acoustid_thread.stop()

        # pool.stop can block a short time because the CV might be locked
        # during starting the pipeline -> idle_add -> no GUI blocking
        GLib.idle_add(idle_cancel)

    def __submit_cb(self, *args):
        self.__submit.set_sensitive(False)
        self.__label.set_markup(util.bold(_("Submitting fingerprints:")))
        self.__set_fraction(0)
        self.__acoustid_thread = AcoustidSubmissionThread(
            list(filter(can_submit, self.__fp_results.values())),
            self.__acoustid_update,
            self.__acoustid_done,
        )

    def __acoustid_update(self, progress):
        self.__set_fraction(progress)
        self.__label_song.set_text(_("Submitting…"))

    def __acoustid_done(self):
        self.__acoustid_thread.join()
        self.__set_fraction(1.0)
        GLib.timeout_add(500, self.destroy)



# Copyright 2012 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import sys
import os

if sys.platform == "win32":
    # To prevent us loading DLLs in the system directory which clash
    # with the ones we ship.
    # https://github.com/quodlibet/quodlibet/issues/2817
    from ctypes import windll

    windll.kernel32.SetDllDirectoryW(os.path.dirname(sys.executable))

from ._import import install_redirect_import_hook

install_redirect_import_hook()

from .util.i18n import _, C_, N_, ngettext, npgettext
from .util.dprint import print_d, print_e, print_w
from ._init import init_cli, init
from ._main import (
    get_base_dir,
    is_release,
    get_user_dir,
    app,
    set_application_info,
    init_plugins,
    enable_periodic_save,
    run,
    finish_first_session,
    get_image_dir,
    is_first_session,
    get_build_description,
    get_build_version,
    get_cache_dir,
)


(
    _,
    C_,
    N_,
    ngettext,
    npgettext,
    is_release,
    init,
    init_cli,
    print_e,
    get_base_dir,
    print_w,
    print_d,
    get_user_dir,
    app,
    set_application_info,
    init_plugins,
    enable_periodic_save,
    run,
    finish_first_session,
    get_image_dir,
    is_first_session,
    get_build_description,
    get_build_version,
    get_cache_dir,
)



# Copyright 2005 Joe Wreschnig
#           2012 Christoph Reiter
#           2014 Jan Path
#      2011-2023 Nick Boultbee
#           2018 David Morris
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from collections.abc import Sequence

from gi.repository import Gtk, GLib, Gdk, GObject
from senf import uri2fsn

from quodlibet import app, print_w, print_d
from quodlibet import config
from quodlibet import const
from quodlibet import qltk
from quodlibet import util
from quodlibet import _

from quodlibet.query import Query
from quodlibet.pattern import Pattern
from quodlibet.qltk.information import Information
from quodlibet.qltk.properties import SongProperties
from quodlibet.qltk.views import AllTreeView, DragScroll
from quodlibet.qltk.ratingsmenu import ConfirmRateMultipleDialog
from quodlibet.qltk.songmodel import PlaylistModel
from quodlibet.qltk import Icons
from quodlibet.qltk.util import GSignals
from quodlibet.qltk.delete import trash_songs
from quodlibet.formats._audio import TAG_TO_SORT, AudioFile
from quodlibet.qltk.x import SeparatorMenuItem
from quodlibet.qltk.songlistcolumns import create_songlist_column, SongListColumn
from quodlibet.util import connect_destroy

DND_QL, DND_URI_LIST = range(2)


class SongSelectionInfo(GObject.Object):
    """
    Songs which get included in the status bar summary.

    The `changed` signal gets fired after any of the songs in the
    selection or the selection itself has changed.
    The signal is async.

    Two selection states:
        - 0 or 1 selected row: all rows
        - 2 or more: only the selected rows

    The signals fires if the state changes.

    FIXME: `row-changed` for song lists isn't implemented (performance).
            Since a library change could change the selection it should
            also trigger this.

            Since this would happen quite often (song stat changes) and
            would lead to a complete re-calc in the common case,
            ignore it for now.
    """

    __gsignals__: GSignals = {
        # changed(songs:list)
        "changed": (GObject.SignalFlags.RUN_LAST, None, (object,))
    }

    def __init__(self, songlist):
        super().__init__()

        self.__idle = None
        self.__songlist = songlist
        self.__selection = sel = songlist.get_selection()
        self.__count = sel.count_selected_rows()
        self.__sel_id = songlist.connect("selection-changed", self.__selection_changed)
        self.__sel_id = songlist.connect("songs-removed", self.__songs_removed)

    def destroy(self):
        self.__songlist.disconnect(self.__sel_id)
        if self.__idle:
            GLib.source_remove(self.__idle)

    def refresh(self):
        songlist = self.__songlist
        songs = songlist.get_selected_songs() or songlist.get_songs()
        self.emit("changed", songs)

    def _update_songs(self, songs):
        """After making changes (filling the list) call this to
        skip any queued changes and emit the passed songs instead"""
        self.__emit_info_selection(songs)
        self.__count = len(songs)

    def __idle_emit(self, songs):
        if songs is None:
            if self.__count <= 1:
                songs = self.__songlist.get_songs()
            else:
                songs = self.__songlist.get_selected_songs()
        self.emit("changed", songs)
        self.__idle = None

    def __emit_info_selection(self, songs=None):
        if self.__idle:
            GLib.source_remove(self.__idle)
        self.__idle = GLib.idle_add(self.__idle_emit, songs, priority=GLib.PRIORITY_LOW)

    def __songs_removed(self, songlist, removed):
        try:
            self.__emit_info_selection()
        except Exception as e:
            print_w(f"Couldn't process removed songs ({e})r")

    def __selection_changed(self, songlist, selection):
        count = selection.count_selected_rows()
        if self.__count == count == 0:
            return
        if count <= 1:
            if self.__count > 1:
                self.__emit_info_selection()
        else:
            self.__emit_info_selection()
        self.__count = count


def get_columns():
    """Gets the list of songlist column headings"""

    columns = config.getstringlist("settings", "columns", const.DEFAULT_COLUMNS)
    if "~current" in columns:
        columns.remove("~current")
    return columns


def set_columns(vals: list[str]) -> None:
    """Persists the settings for songlist headings held in `vals`"""

    config.setstringlist("settings", "columns", vals)


def get_sort_tag(tag):
    """Returns a tag that can be used for sorting for the given column tag.

    Returns '' if the default sort key should be used.
    """

    replace_order = {
        "~#track": "",
        "~#disc": "",
        "~#tracks": "",
        "~#discs": "",
        "~length": "~#length",
    }

    if tag == "~title~version":
        tag = "title"
    elif tag == "~album~discsubtitle":
        tag = "album"

    if "<" in tag:
        for key, value in replace_order.items():
            if value:
                value = f"<{value}>"
            tag = tag.replace(f"<{key}>", value)
        for key, value in TAG_TO_SORT.items():
            tag = tag.replace(f"<{key}>", f"<{value}|<{value}>|<{key}>>")
        tag = Pattern(tag).format
    else:
        tags = util.tagsplit(tag)
        sort_tags = []
        for tag in tags:
            tag = replace_order.get(tag, tag)
            tag = TAG_TO_SORT.get(tag, tag)
            if tag not in sort_tags:
                sort_tags.append(tag)
        if len(sort_tags) > 1:
            tag = "~" + "~".join(sort_tags)

    return tag


def header_tag_split(header):
    """Split a pattern or a tied tag into separate tags"""

    if "<" in header:
        try:
            return list(Pattern(header).tags)
        except ValueError:
            return []
    else:
        return util.tagsplit(header)


class SongListDnDMixin(GObject.GObject):
    """DnD support for the SongList class"""

    def setup_drop(self, library):
        self.connect("drag-begin", self.__drag_begin)
        self.connect("drag-motion", self.__drag_motion)
        self.connect("drag-leave", self.__drag_leave)
        self.connect("drag-data-get", self.__drag_data_get)
        self.connect("drag-data-received", self.__drag_data_received, library)

    def __drag_begin(self, *args):
        ok, state = Gtk.get_current_event_state()
        if ok and state & qltk.get_primary_accel_mod():
            self.__force_copy = True
        else:
            self.__force_copy = False

    def enable_drop(self, by_row=True):
        targets = [
            ("text/x-quodlibet-songs", Gtk.TargetFlags.SAME_APP, DND_QL),
            ("text/uri-list", 0, DND_URI_LIST),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]
        self.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            targets,
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )
        self.drag_dest_set(
            Gtk.DestDefaults.ALL, targets, Gdk.DragAction.COPY | Gdk.DragAction.MOVE
        )
        self.__drop_by_row = by_row
        self.__force_copy = False

    def disable_drop(self):
        targets = [
            ("text/x-quodlibet-songs", Gtk.TargetFlags.SAME_APP, DND_QL),
            ("text/uri-list", 0, DND_URI_LIST),
        ]
        targets = [Gtk.TargetEntry.new(*t) for t in targets]
        self.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, targets, Gdk.DragAction.COPY
        )
        self.drag_dest_unset()

    def __drag_motion(self, view, ctx, x, y, time):
        if self.__drop_by_row:
            self.set_drag_dest(x, y)
            self.scroll_motion(x, y)
            if Gtk.drag_get_source_widget(ctx) == self and not self.__force_copy:
                kind = Gdk.DragAction.MOVE
            else:
                kind = Gdk.DragAction.COPY
            Gdk.drag_status(ctx, kind, time)
            return True
        self.get_parent().drag_highlight()
        Gdk.drag_status(ctx, Gdk.DragAction.COPY, time)
        return True

    def __drag_leave(self, widget, ctx, time):
        widget.get_parent().drag_unhighlight()
        self.scroll_disable()

    def __drag_data_get(self, view, ctx, sel, tid, etime):
        model, paths = self.get_selection().get_selected_rows()
        if tid == DND_QL:
            songs = [model[path][0] for path in paths if model[path][0].can_add]
            if len(songs) != len(paths):
                qltk.ErrorMessage(
                    qltk.get_top_parent(self),
                    _("Unable to copy songs"),
                    _(
                        "The files selected cannot be copied to other "
                        "song lists or the queue."
                    ),
                ).run()
                Gdk.drag_abort(ctx, etime)
                return

            qltk.selection_set_songs(sel, songs)

            # DEM 2018/05/25: The below check is a deliberate repetition of
            # code in the drag-motion signal handler.  In MacOS/Quartz, the
            # context action is not propogated between event handlers for
            # drag-motion and drag-data-get using "ctx.get_actions()".  It is
            # unclear if this is a bug or expected behavior.  Regardless, the
            # context widget information is the same so identical behavior can
            # be achieved by simply using the same widget check as in the move
            # action.
            if Gtk.drag_get_source_widget(ctx) == self and not self.__force_copy:
                self.__drag_iters = list(map(model.get_iter, paths))
            else:
                self.__drag_iters = []
        else:
            uris = [model[path][0]("~uri") for path in paths]
            sel.set_uris(uris)
            self.__drag_iters = []

    def __drag_data_received(self, view, ctx, x, y, sel, info, etime, library):
        model = view.get_model()
        if info == DND_QL:
            filenames = qltk.selection_get_filenames(sel)
            move = bool(ctx.get_selected_action() & Gdk.DragAction.MOVE)
        elif info == DND_URI_LIST:

            def to_filename(s):
                try:
                    return uri2fsn(s)
                except ValueError:
                    return None

            filenames = list(filter(None, map(to_filename, sel.get_uris())))
            move = False
        else:
            Gtk.drag_finish(ctx, False, False, etime)
            return
        # Should always have one here, but you never know (also: types)
        librarian = library.librarian or library
        to_add = []
        for filename in filenames:
            if filename not in librarian:
                library.add_filename(filename)
            elif filename not in library:
                to_add.append(librarian[filename])
        library.add(to_add)
        songs: list = list(filter(None, map(library.get, filenames)))
        if not songs:
            Gtk.drag_finish(ctx, bool(not filenames), False, etime)
            return

        if not self.__drop_by_row:
            success = self.__drag_data_browser_dropped(songs)
            Gtk.drag_finish(ctx, success, False, etime)
            return

        try:
            path, position = view.get_dest_row_at_pos(x, y)
        except TypeError:
            path = max(0, len(model) - 1)
            position = Gtk.TreeViewDropPosition.AFTER

        if move and Gtk.drag_get_source_widget(ctx) == view:
            iter = model.get_iter(path)  # model can't be empty, we're moving
            if position in (
                Gtk.TreeViewDropPosition.BEFORE,
                Gtk.TreeViewDropPosition.INTO_OR_BEFORE,
            ):
                while self.__drag_iters:
                    model.move_before(self.__drag_iters.pop(0), iter)
            else:
                while self.__drag_iters:
                    model.move_after(self.__drag_iters.pop(), iter)
            Gtk.drag_finish(ctx, True, False, etime)
        else:
            song = songs.pop(0)
            try:
                iter = model.get_iter(path)
            except ValueError:
                iter = model.append(row=[song])  # empty model
            else:
                if position in (
                    Gtk.TreeViewDropPosition.BEFORE,
                    Gtk.TreeViewDropPosition.INTO_OR_BEFORE,
                ):
                    iter = model.insert_before(iter, [song])
                else:
                    iter = model.insert_after(iter, [song])
            for song in songs:
                iter = model.insert_after(iter, [song])
            Gtk.drag_finish(ctx, True, move, etime)

    def __drag_data_browser_dropped(self, songs):
        window = qltk.get_top_parent(self)
        return window.browser.dropped(songs)


class SongList(AllTreeView, SongListDnDMixin, DragScroll, util.InstanceTracker):
    """A TreeView containing a list of songs."""

    __gsignals__: GSignals = {
        "songs-removed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        "orders-changed": (GObject.SignalFlags.RUN_LAST, None, []),
    }

    headers: list[str] = []
    """The list of current headers."""

    star = list(Query.STAR)

    def menu(self, header: str, browser, library):
        songs = self.get_selected_songs()
        if not songs:
            return None

        def Filter(t):
            # Translators: The substituted string is the name of the
            # selected column (a translated tag name).
            b = qltk.MenuItem(_("_Filter on %s") % util.tag(t, True), Icons.EDIT_FIND)
            b.connect("activate", self.__filter_on, t, songs, browser)
            return b

        header = header_tag_split(header)[0]
        can_filter = browser.can_filter
        menu_items = []
        if header not in ["artist", "album"] and can_filter(header):
            menu_items.append(Filter(header))
        if can_filter("artist"):
            menu_items.append(Filter("artist"))
        if can_filter("album"):
            menu_items.append(Filter("album"))

        menu = browser.menu(songs, library, items=[menu_items])
        menu.show_all()
        return menu

    def __init__(
        self,
        library,
        player=None,
        update=False,
        model_cls=PlaylistModel,
        sortable: bool = True,
    ):
        super().__init__()
        self.sortable = sortable
        self._register_instance(SongList)
        self.set_model(model_cls())
        self.info = SongSelectionInfo(self)
        self.set_size_request(200, 150)
        self.set_rules_hint(True)
        self.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        self.set_fixed_height_mode(True)
        self.__csig = self.connect("columns-changed", self.__columns_changed)
        self._first_column = None
        # A priority list of how to apply the sort keys.
        # might contain column header names not present...
        self._sort_sequence: list[str] = []
        self.set_column_headers(self.headers)
        librarian = library.librarian or library

        connect_destroy(librarian, "changed", self.__song_updated)
        connect_destroy(librarian, "removed", self.__song_removed, player)

        if update:
            connect_destroy(librarian, "added", self.__song_added)

        if player:
            connect_destroy(player, "paused", lambda *x: self.__redraw_current())
            connect_destroy(player, "unpaused", lambda *x: self.__redraw_current())
            connect_destroy(player, "error", lambda *x: self.__redraw_current())

        self.connect("button-press-event", self.__button_press, library)
        self.connect("key-press-event", self.__key_press, library, player)

        self.setup_drop(library)
        self.disable_drop()

        self.set_search_equal_func(self.__search_func, None)

        self.connect("destroy", self.__destroy)

    @property
    def sortable(self) -> bool:
        """Whether the columns clicked to enable sorting of songs"""
        return self._sortable

    @sortable.setter
    def sortable(self, value: bool):
        # It's either sortable or clickable columns, both ends in a buggy UI (see #4099)
        always_sortable = config.getboolean("song_list", "always_allow_sorting")
        self._sortable = value or always_sortable
        for col in self.get_columns():
            try:
                col.get_widget().set_sensitive(self._sortable)
            except (TypeError, AttributeError):
                # Just in case columns don't always work like this
                pass

    @property
    def model(self) -> Gtk.TreeModel:
        return self.get_model()

    @property
    def sourced(self):
        return self.model.sourced

    def toggle_column_sort(self, column, replace=True, refresh=True):
        """Toggles the sort order of a column.

        If not sorted, defaults to Gtk.SortType.ASCENDING

        If replace is False, the column will be appended to existing
        sorted columns. If it replaces a sort sequence where it was part of
        before it will not toggle itself, only remove the other ones.

        If refresh is True, the song list will be resorted.
        """

        if not self.sortable:
            return

        # update the sort priority list
        if replace:
            del self._sort_sequence[:]
        else:
            try:
                self._sort_sequence.remove(column.header_name)
            except ValueError:
                pass
        self._sort_sequence.append(column.header_name)

        # in case we replace a multi sort with one sort that was part before
        # don't toggle, because it usually means we want to get rid of
        # the other one
        dont_reverse = False
        if replace and column.get_sort_indicator():
            for c in self.get_columns():
                if c is not column and c.get_sort_indicator():
                    dont_reverse = True
                    break

        # set the indicators
        default_order = Gtk.SortType.ASCENDING
        for c in self.get_columns():
            if c is column:
                if c.get_sort_indicator():
                    if dont_reverse:
                        order = c.get_sort_order()
                    else:
                        order = not c.get_sort_order()
                else:
                    order = default_order
                c.set_sort_order(order)
                c.set_sort_indicator(True)
                if not replace:
                    break
            elif replace:
                c.set_sort_indicator(False)

        if refresh:
            songs = self.get_songs()
            song_order = self._get_song_order(songs)
            self.model.reorder(song_order)

            selection = self.get_selection()
            _, paths = selection.get_selected_rows()
            if paths:
                self.scroll_to_cell(paths[0], use_align=True, row_align=0.5)

        self.emit("orders-changed")

    def find_default_sort_column(self):
        """Returns a column that will sort using only the default sort key
        or None if none can't be found
        """

        for c in self.get_columns():
            # get_sort_tag == "" if the default sort key should be used
            if not get_sort_tag(c.header_name):
                return c
        return None

    def is_sorted(self):
        """If any of the columns has a sort indicator.

        This does not mean that the list content is sorted.
        """

        for c in self.get_columns():
            if c.get_sort_indicator():
                return True
        return False

    def clear_sort(self):
        """Remove all column sort indicators"""

        for h in self.get_columns():
            h.set_sort_indicator(False)
        del self._sort_sequence[:]

        self.emit("orders-changed")

    def get_sort_orders(self):
        """Returns a list of tuples (header_name, descending)"""

        sorted_ = [c for c in self.get_columns() if c.get_sort_indicator()]

        # if someone adds columns and sorts them using the TV API directly..
        # better not crash I guess
        for c in sorted_:
            if c.header_name not in self._sort_sequence:
                self._sort_sequence.append(c.header_name)

        sorted_.sort(key=lambda c: self._sort_sequence.index(c.header_name))
        return [(c.header_name, bool(c.get_sort_order())) for c in sorted_]

    def set_sort_orders(self, orders):
        """Pass a value returned by get_sort_orders() to restore the state"""

        if not self.sortable:
            return

        self._sort_sequence = [tag for tag, o in orders]

        orders = dict(orders)
        for c in self.get_columns():
            if c.header_name in orders:
                c.set_sort_indicator(True)
                c.set_sort_order(orders[c.header_name])
            else:
                c.set_sort_indicator(False)

        self.emit("orders-changed")

    def __destroy(self, *args):
        self.info.destroy()
        self.info = None
        self.handler_block(self.__csig)
        for column in self.get_columns():
            self.remove_column(column)
        self.handler_unblock(self.__csig)

    def __search_func(self, model, column, key, iter, *args):
        for column in self.get_columns():
            value = model.get_value(iter)(column.header_name)
            if not isinstance(value, str):
                continue
            if key in value.lower() or key in value:
                return False
        else:
            return True

    def __filter_on(self, widget, header, songs, browser):
        if not browser:
            return

        # Fall back to the playing song
        if songs is None:
            if app.player.song:
                songs = [app.player.song]
            else:
                return

        browser.filter_on(songs, header)

    def __button_press(self, view, event, librarian):
        if event.button != Gdk.BUTTON_PRIMARY:
            return None
        x, y = map(int, [event.x, event.y])
        try:
            path, col, cellx, celly = view.get_path_at_pos(x, y)
        except TypeError:
            return True
        if event.window != self.get_bin_window():
            return False
        if col.header_name == "~rating":
            if not config.getboolean("browsers", "rating_click"):
                return None

            song = view.get_model()[path][0]
            l = Gtk.Label()
            l.show()
            l.set_text(config.RATINGS.full_symbol)
            width = l.get_preferred_size()[1].width
            l.destroy()
            if not width:
                return False
            precision = config.RATINGS.precision
            count = int(float(cellx - 5) / width) + 1
            rating = max(0.0, min(1.0, count * precision))
            if rating <= precision and song("~#rating") == precision:
                rating = 0.0
            self.__set_rating(rating, [song], librarian)
        return None

    def __set_rating(self, value, songs, librarian):
        count = len(songs)
        if count > 1 and config.getboolean("browsers", "rating_confirm_multiple"):
            dialog = ConfirmRateMultipleDialog(self, count, value)
            if dialog.run() != Gtk.ResponseType.YES:
                return
        for song in songs:
            song["~#rating"] = value
        librarian.changed(songs)

    def __key_press(self, songlist, event, librarian, player):
        if qltk.is_accel(event, "<Primary>Return", "<Primary>KP_Enter"):
            self.__enqueue(self.get_selected_songs())
            return True
        if qltk.is_accel(event, "<Primary>F"):
            self.emit("start-interactive-search")
            return True
        if qltk.is_accel(event, "<Primary>Delete"):
            songs = self.get_selected_songs()
            if songs:
                trash_songs(self, songs, librarian)
            return True
        if qltk.is_accel(event, "<alt>Return"):
            songs = self.get_selected_songs()
            if songs:
                window = SongProperties(librarian, songs, parent=self)
                window.show()
            return True
        if qltk.is_accel(event, "<Primary>I"):
            songs = self.get_selected_songs()
            if songs:
                window = Information(librarian, songs, self)
                window.show()
            return True
        if qltk.is_accel(event, "space", "KP_Space") and player is not None:
            player.paused = not player.paused
            return True
        if qltk.is_accel(event, "F2"):
            songs = self.get_selected_songs()
            if len(songs) > 1:
                print_d("Can't edit more than one")
            elif songs:
                path, col = songlist.get_cursor()
                song = self.get_first_selected_song()
                cls = type(col).__name__
                if col.can_edit:
                    print_d(f"Let's edit this: {song} ({cls} can be edited)")
                    renderers = col.get_cells()
                    renderers[0].props.editable = True
                    self.set_cursor(path, col, start_editing=True)
                else:
                    print_d(f"Can't edit {cls}. Maybe it's synthetic / numeric?")

        return False

    def __enqueue(self, songs):
        songs = [s for s in songs if s.can_add]
        if songs:
            from quodlibet import app

            app.window.playlist.enqueue(songs)

    def __redraw_current(self):
        model = self.get_model()
        iter_ = model.current_iter
        if iter_:
            path = model.get_path(iter_)
            model.row_changed(path, iter_)

    def __columns_changed(self, *args):
        headers = [h.header_name for h in self.get_columns()]
        SongList.set_all_column_headers(headers)
        SongList.headers = headers

    def __column_width_changed(self, *args):
        # make sure non-resizable columns stay non-expanding.
        # gtk likes to change them sometimes
        for c in self.get_columns():
            if not c.get_resizable() and c.get_expand():
                c.set_expand(False)

        widths = []
        expands = []
        for c in self.get_columns():
            if not c.get_resizable():
                continue
            widths.extend((c.header_name, str(c.get_fixed_width())))
            expands.extend((c.header_name, str(int(c.get_expand()))))
        config.setstringlist("memory", "column_widths", widths)
        config.setstringlist("memory", "column_expands", expands)

    @classmethod
    def set_all_column_headers(cls, headers: list[str]):
        set_columns(headers)
        try:
            headers.remove("~current")
        except ValueError:
            pass
        cls.headers = headers
        for listview in cls.instances():
            listview.set_column_headers(headers)

        star = list(Query.STAR)
        for header in headers:
            for tag in header_tag_split(header):
                if not tag.startswith("~#") and tag not in star:
                    star.append(tag)

        for tag in config.getlist("settings", "search_tags"):
            if tag and tag not in star:
                star.append(tag)

        SongList.star = star

    def set_model(self, model):
        super().set_model(model)
        self.set_search_column(0)

    def clear(self):
        """Remove all songs"""

        model = self.get_model()
        if model:
            model.clear()

    def get_songs(self):
        """Get all songs currently in the song list"""

        model = self.get_model()
        if not model:
            return []
        return model.get()

    def _get_song_order(self, songs: list[AudioFile]) -> Sequence[int] | None:
        """Returns mapping from new position to position in given list of songs
        when sorted based on the column sort orders"""

        orders = self.get_sort_orders()
        if orders:
            song_order = list(range(len(songs)))
            for key, reverse in self.__get_song_sort_key_func(orders):
                song_order.sort(key=lambda i: key(songs[i]), reverse=reverse)
            return song_order
        return None

    def __get_song_sort_key_func(self, order):
        last_tag = None
        last_order = None
        first = True
        key_func = []
        for tag, reverse in order:
            tag = get_sort_tag(tag)

            # always sort using the default sort key first
            if first:
                first = False
                key_func.append((lambda s: s.sort_key, reverse))
                last_order = reverse
                last_tag = ""

            # no need to sort twice in a row with the same key/order
            if tag == last_tag and last_order == reverse:
                continue
            last_order = reverse
            last_tag = tag

            if tag == "":
                key_func.append((lambda s: s.sort_key, reverse))
            else:
                sort_func = AudioFile.sort_by_func(tag)
                key_func.append((sort_func, reverse))
        return key_func

    def add_songs(self, songs):
        """Add songs to the list in the right order and position"""

        if not songs:
            return

        model = self.get_model()
        if not len(model):
            self.set_songs(songs, scroll=False)
            return

        if not self.is_sorted():
            model.append_many(songs)
            return

        for song in songs:
            insert_iter = self.__find_song_position(song)
            model.insert_before(insert_iter, row=[song])

    def set_songs(
        self,
        songs: list[AudioFile],
        sorted: bool = False,
        scroll: bool = True,
        scroll_select: bool = False,
    ):
        """Fill the song list.

        If sorted is True, the passed songs will not be sorted and
        all sort indicators will be removed.

        If scroll is True the list will scroll to the current song.

        If scroll_select is True restore the selection of the first
        selected song and scroll to. Falls back to the current song.
        """

        model = self.get_model()
        assert model is not None

        song_order = None

        if not sorted:
            # make sure some sorting is set and visible
            if not self.is_sorted():
                default = self.find_default_sort_column()
                if default:
                    self.toggle_column_sort(default, refresh=False)
            song_order = self._get_song_order(songs)
        else:
            self.clear_sort()

        restore_song = None
        if scroll_select:
            restore_song = self.get_first_selected_song()

        with self.without_model() as model:
            model.set(songs)
            if song_order:
                model.reorder(song_order)

        # scroll to the first selected or current song and restore
        # selection for the first selected item if there was one
        if scroll or scroll_select:
            if restore_song is not None and restore_song is not model.current:
                try:
                    index = songs.index(restore_song)
                except ValueError:
                    path = model.current_path
                else:
                    path = Gtk.TreePath.new()
                    path.append_index(index)
            else:
                path = model.current_path

            if path is not None:
                if restore_song is not None:
                    self.set_cursor(path)
                self.scroll_to_cell(path, use_align=True, row_align=0.5)

        # the song selection has queued a change now, cancel that and
        # pass the songs manually
        self.info._update_songs(songs)

    def jump_to_song(self, song, select=False):
        """Scrolls to and selects the given song if in the list.

        Returns True if the song was found.
        """

        model = self.get_model()
        if not model:
            return False

        # fast path
        if song == model.current:
            path = model.current_path
        else:
            iter_ = model.find(song)
            if iter_ is None:
                return False
            path = model.get_path(iter_)

        if select:
            self.set_cursor(path)
        self.scroll_to_cell(path, use_align=True, row_align=0.5)

        return True

    def get_first_selected_song(self):
        """The first selected song in the list or None"""

        selection = self.get_selection()
        model, paths = selection.get_selected_rows()
        if paths:
            return model.get_value(model.get_iter(paths[0]))
        return None

    def get_selected_songs(self):
        """Returns a list of selected songs"""

        songs = []

        def func(model, path, iter_, user_data):
            songs.append(model.get_value(iter_))

        selection = self.get_selection()
        selection.selected_foreach(func, None)
        return songs

    def __find_song_position(self, song):
        """Finds the appropriate position of a song in a sorted song list.

        Returns iter of the song after the given song according to the current
        sort order.

        Returns None if the correct position is at the end of the song list.
        """

        model = self.get_model()
        order = self.get_sort_orders()
        sort_key_func = list(enumerate(reversed(self.__get_song_sort_key_func(order))))
        song_sort_keys = [key(song) for i, (key, r) in sort_key_func]
        i = 0
        j = len(model)
        while i < j:
            mid = (i + j) // 2
            other_song_iter = model.iter_nth_child(None, mid)
            other_song = model.get_value(other_song_iter)
            song_is_lower = False
            for i, (key, reverse) in sort_key_func:
                other_key = key(other_song)
                is_lower = song_sort_keys[i] < other_key
                is_greater = song_sort_keys[i] > other_key
                if not reverse and is_lower or reverse and is_greater:
                    song_is_lower = True
                    break
                if not reverse and is_greater or reverse and is_lower:
                    break
            if song_is_lower:
                j = mid
            else:
                i = mid + 1
        if i < len(model):
            return model.iter_nth_child(None, i)
        return None

    def __find_iters_in_selection(self, songs) -> tuple[list, list, bool]:
        model, rows = self.get_selection().get_selected_rows()
        rows = rows or []
        iters = []
        removed_songs = []
        for r in rows:
            if model[r][0] in songs:
                iters.append(model[r].iter)
                removed_songs.append(model[r][0])
        complete = len(iters) == len(songs)
        return iters, removed_songs, complete

    def __song_updated(self, librarian, songs):
        """Only update rows that are currently displayed.
        Warning: This makes the row-changed signal useless.
        """
        model = self.get_model()
        if config.getboolean("song_list", "auto_sort") and self.is_sorted():
            iters, _, complete = self.__find_iters_in_selection(songs)

            if not complete:
                iters = model.find_all(songs)

            rows = [Gtk.TreeRowReference.new(model, model.get_path(i)) for i in iters]

            for row in rows:
                iter = model.get_iter(row.get_path())
                song = model.get_value(iter)
                insert_iter = self.__find_song_position(song)
                model.move_before(iter, insert_iter)

        vrange = self.get_visible_range()
        if vrange is None:
            return
        (start,), (end,) = vrange
        for path in range(start, end + 1):
            row = model[path]
            if row[0] in songs:
                model.row_changed(row.path, row.iter)

    def __song_added(self, librarian, songs):
        window = qltk.get_top_parent(self)
        filter_ = window.browser.active_filter
        if callable(filter_):
            self.add_songs(list(filter(filter_, songs)))

    def __song_removed(self, librarian, songs, player):
        try:
            # The player needs to be called first so it can ge the next song
            # in case the current one gets deleted and the order gets reset.
            if player:
                for song in songs:
                    player.remove(song)

            model = self.get_model()

            # The selected songs are removed from the library and should
            # be removed from the view.

            if not len(model):
                return

            songs = set(songs)

            # search in the selection first
            # speeds up common case: select songs and remove them
            iters, removed_songs, complete = self.__find_iters_in_selection(songs)

            # if not all songs were in the selection, search the whole view
            if not complete:
                removed_songs = []
                for iter_, value in self.model.iterrows():
                    if value in songs:
                        iters.append(iter_)
                        removed_songs.append(value)

            if removed_songs:
                self.emit("songs-removed", set(removed_songs))
            self.remove_iters(iters)
        except Exception as e:
            print_w(f"Couldn't process removed songs: {e}", self)

    def set_first_column_type(self, column_type):
        """Set a column that will be included at the beginning"""

        self._first_column = column_type

        # refresh
        self.set_column_headers(self.headers)

    # Build a new filter around our list model, set the headers to their
    # new values.
    def set_column_headers(self, headers):
        if len(headers) == 0:
            return

        self.handler_block(self.__csig)

        old_sort = self.get_sort_orders()
        for column in self.get_columns():
            self.remove_column(column)

        if self._first_column:
            column = self._first_column()
            self.append_column(column)

        cws = config.getstringlist("memory", "column_widths")
        column_widths = {}
        for i in range(0, len(cws), 2):
            column_widths[cws[i]] = int(cws[i + 1])

        ce = config.getstringlist("memory", "column_expands")
        column_expands = {}
        for i in range(0, len(ce), 2):
            column_expands[ce[i]] = int(ce[i + 1])

        for t in headers:
            column = create_songlist_column(self.model, t)
            if column.get_resizable():
                if t in column_widths:
                    column.set_fixed_width(column_widths[t])
                if t in column_expands:
                    column.set_expand(column_expands[t])
                else:
                    column.set_expand(True)

            def column_clicked(column, *args):
                if not self._sortable:
                    return
                # if ctrl is held during the sort click, append a sort key
                # or change order if already sorted
                ctrl_held = False
                event = Gtk.get_current_event()
                if event:
                    ok, state = event.get_state()
                    if ok and state & qltk.get_primary_accel_mod():
                        ctrl_held = True

                self.toggle_column_sort(column, replace=not ctrl_held)

            column.connect("clicked", column_clicked)
            column.connect("button-press-event", self.__showmenu)
            column.connect("popup-menu", self.__showmenu)
            column.connect("notify::width", self.__column_width_changed)
            column.set_reorderable(True)
            self.append_column(column)

        self.set_sort_orders(old_sort)
        self.columns_autosize()

        self.handler_unblock(self.__csig)

    def _menu(self, column: SongListColumn) -> Gtk.Menu:
        menu = Gtk.Menu()

        def selection_done_cb(menu):
            menu.destroy()

        menu.connect("selection-done", selection_done_cb)

        current_set = set(SongList.headers)

        def tag_title(tag: str):
            if "<" in tag:
                return util.pattern(tag)
            return util.tag(tag)

        current = [(tag_title(c), c) for c in SongList.headers]

        def add_header_toggle(
            menu: Gtk.Menu,
            header: str,
            tag: str,
            active: bool,
            column: SongListColumn = column,
        ):
            item = Gtk.CheckMenuItem(label=header)
            item.tag = tag
            item.set_active(active)
            item.connect("activate", self.__toggle_header_item, column)
            item.show()
            item.set_tooltip_text(tag)
            menu.append(item)

        for header, tag in current:
            add_header_toggle(menu, header, tag, True)

        sep = SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        trackinfo = """title genre comment ~title~version ~#track
            ~#playcount ~#skipcount ~rating ~#length ~playlists
            bpm initialkey""".split()
        peopleinfo = """artist ~people performer arranger author composer
            conductor lyricist originalartist""".split()
        albuminfo = """album ~album~discsubtitle labelid ~#disc ~#discs
            ~#tracks albumartist""".split()
        dateinfo = """date originaldate recordingdate ~year ~originalyear
            ~#laststarted ~#lastplayed ~#added ~#mtime""".split()
        fileinfo = """~format ~#bitdepth ~#bitrate ~#filesize
            ~filename ~basename ~dirname ~uri ~codec ~encoding ~#channels
            ~#samplerate""".split()
        copyinfo = """copyright organization location isrc
            contact website""".split()
        all_headers: list[str] = sum(
            [trackinfo, peopleinfo, albuminfo, dateinfo, fileinfo, copyinfo], []
        )

        for name, group in [
            (_("All _Headers"), all_headers),
            (_("_Track Headers"), trackinfo),
            (_("_Album Headers"), albuminfo),
            (_("_People Headers"), peopleinfo),
            (_("_Date Headers"), dateinfo),
            (_("_File Headers"), fileinfo),
            (_("_Production Headers"), copyinfo),
        ]:
            item = Gtk.MenuItem(label=name, use_underline=True)
            item.show()
            menu.append(item)
            submenu = Gtk.Menu()
            item.set_submenu(submenu)
            for header, tag in sorted(zip(map(util.tag, group), group)):  # noqa
                add_header_toggle(submenu, header, tag, tag in current_set)

        sep = SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        custom = Gtk.MenuItem(label=_("_Customize Headers…"), use_underline=True)
        custom.show()
        custom.connect("activate", self.__add_custom_column)
        menu.append(custom)

        item = Gtk.CheckMenuItem(label=_("_Expand Column"), use_underline=True)
        item.set_active(column.get_expand())
        item.set_sensitive(column.get_resizable())

        def set_expand_cb(item, column):
            do_expand = item.get_active()
            if not do_expand:
                # in case we unexpand, get the current width and set it
                # so the column doesn't give up all its space
                # to the left over expanded columns
                column.set_fixed_width(column.get_width())
            else:
                # in case we expand this seems to trigger a re-distribution
                # between all expanded columns
                column.set_fixed_width(-1)
            column.set_expand(do_expand)
            self.columns_autosize()

        sep = SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        item.connect("activate", set_expand_cb, column)
        item.show()
        menu.append(item)

        return menu

    def __toggle_header_item(self, item, column):
        headers = SongList.headers[:]
        if item.get_active():
            try:
                headers.insert(self.get_columns().index(column), item.tag)
            except ValueError:
                headers.append(item.tag)
        else:
            try:
                headers.remove(item.tag)
            except ValueError:
                pass

        SongList.set_all_column_headers(headers)
        SongList.headers = headers

    def __add_custom_column(self, item):
        # Prefs has to import SongList, so do this here to avoid
        # a circular import.
        from quodlibet.qltk.prefs import PreferencesWindow

        window = PreferencesWindow(self)
        window.show()
        window.set_page("songlist")

    def __showmenu(self, column, event=None):
        time = event.time if event else Gtk.get_current_event_time()

        if event is not None and event.button != Gdk.BUTTON_SECONDARY:
            return False

        menu = self._menu(column)
        menu.attach_to_widget(self, None)

        if event:
            menu.popup(None, None, None, None, event.button, time)
            return True

        widget = column.get_widget()
        qltk.popup_menu_under_widget(menu, widget, 3, time)
        return True


@config.register_upgrade_function
def _migrate_rating_column(config, old, new):
    if old < 0:
        # https://github.com/quodlibet/quodlibet/issues/1381
        columns = get_columns()[:]
        for i, c in enumerate(columns):
            if c == "~#rating":
                columns[i] = "~rating"
        set_columns(columns)



# Copyright 2012-2015 Ryan "ZDBioHazard" Turner <zdbiohazard2@gmail.com>
#             2016-22 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _, util
from quodlibet.plugins.songshelpers import each_song, is_writable
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk import Icons


class EditPlaycount(SongsMenuPlugin):
    PLUGIN_ID = "editplaycount"
    PLUGIN_NAME = _("Edit Playcount")
    PLUGIN_DESC_MARKUP = _(
        "Edit a song's <tt>~#playcount</tt> and <tt>~#skipcount.</tt>"
        "\n\n"
        "When multiple songs are selected, counts will be "
        "incremented, rather than set."
        "\n\n"
        "When setting a song's <tt>~#playcount</tt> to 0, "
        "the <tt>~#lastplayed</tt> and <tt>~#laststarted</tt> "
        "entries will be cleared. "
        "However, when setting a 0-play song to a positive play "
        "count, no play times will be created."
    )
    PLUGIN_ICON = Icons.EDIT
    REQUIRES_ACTION = True

    plugin_handles = each_song(is_writable)

    def plugin_songs(self, songs):
        # This is just here so the spinner has something to call. >.>
        def response(win, response_id):
            dlg.response(response_id)
            return

        # Create a dialog.
        dlg = Gtk.Dialog(
            title=_("Edit Playcount"),
            flags=(Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT),
        )
        dlg.add_button(_("_Cancel"), Gtk.ResponseType.REJECT)
        dlg.add_button(_("_Apply"), Gtk.ResponseType.APPLY)
        dlg.set_default_response(Gtk.ResponseType.APPLY)
        dlg.set_border_width(4)
        dlg.vbox.set_spacing(4)

        # Create some spinners.
        play = Gtk.SpinButton()
        play.set_adjustment(Gtk.Adjustment(0, -1000, 1000, 1, 1))
        skip = Gtk.SpinButton()
        skip.set_adjustment(Gtk.Adjustment(0, -1000, 1000, 1, 1))

        # Connect the signals.
        play.connect("activate", response, Gtk.ResponseType.APPLY)
        skip.connect("activate", response, Gtk.ResponseType.APPLY)

        # Set some defaults.
        play.set_numeric(True)
        skip.set_numeric(True)

        # Put all this stuff in a pretty table.
        table = Gtk.Table(rows=2, columns=2)
        table.set_row_spacings(4)
        table.set_col_spacings(4)
        table.attach(Gtk.Label(_("Play Count")), 0, 1, 0, 1)
        table.attach(Gtk.Label(_("Skip Count")), 0, 1, 1, 2)
        table.attach(play, 1, 2, 0, 1)
        table.attach(skip, 1, 2, 1, 2)
        dlg.vbox.add(table)

        # Make a couple tweaks based on the current mode.
        if len(songs) == 1:
            play.set_adjustment(Gtk.Adjustment(0, 0, 9999, 1, 1))
            skip.set_adjustment(Gtk.Adjustment(0, 0, 9999, 1, 1))
            play.set_value(songs[0].get("~#playcount", 0))
            skip.set_value(songs[0].get("~#skipcount", 0))
        else:
            note = Gtk.Label()
            note.set_justify(Gtk.Justification.CENTER)
            note.set_markup(
                util.bold(_("Multiple files selected."))
                + "\n"
                + _("Counts will be incremented.")
            )
            dlg.vbox.add(note)

        dlg.show_all()

        # Only operate if apply is pressed.
        if dlg.run() == Gtk.ResponseType.APPLY:
            for song in songs:
                # Increment when not in single mode.
                if len(songs) == 1:
                    song["~#playcount"] = play.get_value_as_int()
                    song["~#skipcount"] = skip.get_value_as_int()
                else:  # Can't use += here because these tags might not exist.
                    song["~#playcount"] = max(
                        0, (song.get("~#playcount", 0) + play.get_value_as_int())
                    )
                    song["~#skipcount"] = max(
                        0, (song.get("~#skipcount", 0) + skip.get_value_as_int())
                    )

                # When the playcount is set to 0, delete the playcount
                # itself and the last played/started time. We don't
                # want unused or impossible data floating around.
                if song.get("~#playcount", 0) == 0:
                    for tag in ["~#playcount", "~#lastplayed", "~#laststarted"]:
                        song.pop(tag, None)

                # Also delete the skip count if it's zero.
                if song.get("~#skipcount", 0) == 0:
                    song.pop("~#skipcount", None)

        dlg.destroy()
        return



# Copyright 2013 Simonas Kazlauskas
#      2014-2021 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from itertools import chain

from gi.repository import GObject

from quodlibet import _
from quodlibet.formats import AudioFile
from quodlibet.plugins import PluginManager, PluginHandler
from quodlibet.qltk.notif import Task
from quodlibet.util.cover import built_in
from quodlibet.util import print_d
from quodlibet.util.thread import call_async
from quodlibet.util.thumbnails import get_thumbnail_from_file
from quodlibet.plugins.cover import CoverSourcePlugin


class CoverPluginHandler(PluginHandler):
    """A plugin handler for CoverSourcePlugin implementation"""

    def __init__(self, use_built_in=True):
        self.providers = set()
        if use_built_in:
            self.built_in = {built_in.EmbeddedCover, built_in.FilesystemCover}
        else:
            self.built_in = set()

    def plugin_handle(self, plugin):
        return issubclass(plugin.cls, CoverSourcePlugin)

    def plugin_enable(self, plugin):
        self.providers.add(plugin)
        print_d(f"Registered {plugin.cls.__name__} cover source")

    def plugin_disable(self, plugin):
        self.providers.remove(plugin)
        print_d(f"Unregistered {plugin.cls.__name__} cover source")

    @property
    def sources(self):
        """Yields all active CoverSourcePlugin classes sorted by priority"""

        sources = chain((p.cls for p in self.providers), self.built_in)
        yield from sorted(sources, reverse=True, key=lambda x: x.priority())


class CoverManager(GObject.Object):
    __gsignals__ = {
        # ([AudioFile]), emitted if the cover for any songs might have changed
        "cover-changed": (GObject.SignalFlags.RUN_LAST, None, (object,)),
        # Covers were found for the songs
        "covers-found": (GObject.SignalFlags.RUN_LAST, None, (object, object)),
        # All searches were submitted, and success by provider is sent
        "searches-complete": (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    plugin_handler = None

    def __init__(self, use_built_in=True):
        super().__init__()
        self.plugin_handler = CoverPluginHandler(use_built_in)

    def init_plugins(self):
        """Register the cover sources plugin handler with the global
        plugin manager.
        """

        PluginManager.instance.register_handler(self.plugin_handler)

    @property
    def sources(self):
        return self.plugin_handler.sources

    def cover_changed(self, songs):
        """Notify the world that the artwork for some songs or collections
        containing that songs might have changed (For example a new image was
        added to the folder or a new embedded image was added)

        This will invalidate all caches and will notify others that they have
        to re-fetch the cover and do a display update.
        """

        self.emit("cover-changed", songs)

    def acquire_cover(self, callback, cancellable, song):
        """
        Try to get covers from all cover sources until a cover is found.

        * callback(found, result) is the function which will be called when
        this method completes its job.
        * cancellable – Gio.Cancellable which will interrupt the search.
        The callback won't be called when the operation is cancelled.
        """
        sources = self.sources

        def success(source, result):
            name = source.__class__.__name__
            print_d(f"Successfully got cover from {name}")
            source.disconnect_by_func(success)
            source.disconnect_by_func(failure)
            if not cancellable or not cancellable.is_cancelled():
                callback(True, result)

        def failure(source, msg):
            name = source.__class__.__name__
            print_d(f"Didn't get cover from {name}: {msg}")
            source.disconnect_by_func(success)
            source.disconnect_by_func(failure)
            if not cancellable or not cancellable.is_cancelled():
                run()

        def run():
            try:
                provider = next(sources)(song, cancellable)
            except StopIteration:
                return callback(False, None)  # No cover found

            cover = provider.cover
            if cover:
                name = provider.__class__.__name__
                print_d(f"Found local cover from {name}: {cover}")
                callback(True, cover)
            else:
                provider.connect("fetch-success", success)
                provider.connect("fetch-failure", failure)
                provider.fetch_cover()
            return None

        if not cancellable or not cancellable.is_cancelled():
            run()

    def acquire_cover_sync(self, song, embedded=True, external=True):
        """Gets *cached* cover synchronously.

        As CoverSource fetching functionality is asynchronous it is only
        possible to check for already fetched cover.
        """

        return self.acquire_cover_sync_many([song], embedded, external)

    def acquire_cover_sync_many(self, songs, embedded=True, external=True):
        """Same as acquire_cover_sync but returns a cover for multiple
        images"""

        for plugin in self.sources:
            if not embedded and plugin.embedded:
                continue
            if not external and not plugin.embedded:
                continue

            groups = {}
            for song in songs:
                group = plugin.group_by(song) or ""
                groups.setdefault(group, []).append(song)

            # sort both groups and songs by key, so we always get
            # the same result for the same set of songs
            for _key, group in sorted(groups.items()):
                song = sorted(group, key=lambda s: s.key)[0]
                cover = plugin(song).cover
                if cover:
                    return cover
        return None

    def get_cover(self, song):
        """Returns a cover file object for one song or None.

        Compared to acquire_cover_sync() this respects the prefer_embedded
        setting.
        """

        return self.get_cover_many([song])

    def get_cover_many(self, songs):
        """Returns a cover file object for many songs or None.

        Returns the first found image for a group of songs.
        It tries to return the same cover for the same set of songs.
        """

        return self.acquire_cover_sync_many(songs)

    def get_pixbuf_many(self, songs, width, height):
        """Returns a Pixbuf which fits into the boundary defined by width
        and height or None.

        Uses the thumbnail cache if possible.
        """

        fileobj = self.get_cover_many(songs)
        if fileobj is None:
            return None

        return get_thumbnail_from_file(fileobj, (width, height))

    def get_pixbuf(self, song, width, height):
        """see get_pixbuf_many()"""

        return self.get_pixbuf_many([song], width, height)

    def get_pixbuf_many_async(self, songs, width, height, cancel, callback):
        """Async variant; callback gets called with a pixbuf or not called
        in case of an error. cancel is a Gio.Cancellable.

        The callback will be called in the main loop.
        """

        fileobj = self.get_cover_many(songs)
        if fileobj is None:
            return

        call_async(
            get_thumbnail_from_file, cancel, callback, args=(fileobj, (width, height))
        )

    def search_cover(self, cancellable, songs):
        """Search for all the covers applicable to `songs` across all providers
        Every successful image result emits a 'covers-found' signal
        (unless cancelled)."""

        sources = [source for source in self.sources if not source.embedded]
        processed = {}
        all_groups = {}
        task = Task(
            _("Cover Art"), _("Querying album art providers"), stop=cancellable.cancel
        )

        def finished(provider, success):
            processed[provider] = success
            total = self._total_groupings(all_groups)

            frac = len(processed) / total
            print_d(f"Got result for {provider} ({len(processed)} / {total} group(s))")
            task.update(frac)
            if frac >= 1:
                task.finish()
                self.emit("searches-complete", processed)

        def search_complete(provider, results):
            name = provider.name
            if not results:
                print_d(f"No covers from {name}")
                finished(provider, False)
                return
            finished(provider, True)
            if not (cancellable and cancellable.is_cancelled()):
                covers = {
                    CoverData(
                        url=res["cover"],
                        source=name,
                        dimensions=res.get("dimensions", None),
                    )
                    for res in results
                }
                self.emit("covers-found", provider, covers)
            provider.disconnect_by_func(search_complete)

        def failure(provider, result):
            finished(provider, False)
            name = provider.__class__.__name__
            print_d(f"Failed to get cover from {name!r} ({result})")
            provider.disconnect_by_func(failure)

        def song_groups(songs, sources):
            all_groups = {}
            for plugin in sources:
                groups = {}
                for song in songs:
                    group = plugin.group_by(song) or ""
                    groups.setdefault(group, []).append(song)
                all_groups[plugin] = groups
            return all_groups

        all_groups = song_groups(songs, sources)
        print_d(f"Got {self._total_groupings(all_groups)} plugin groupings")

        for plugin_cls, groups in all_groups.items():
            for key, group in sorted(groups.items()):
                song = sorted(group, key=lambda s: s.key)[0]
                artists = {s.comma("artist") for s in group}
                if len(artists) > 1:
                    print_d(
                        f"{len(artists)} artist groups in {key} "
                        "- probably a compilation. "
                        "Using provider to search for compilation"
                    )
                    song = AudioFile(song)
                    try:
                        del song["artist"]
                    except KeyError:
                        # Artist(s) from other grouped songs, never mind.
                        pass
                provider = plugin_cls(song)
                provider.connect("search-complete", search_complete)
                provider.connect("fetch-failure", failure)
                provider.search()
        return all_groups

    def _total_groupings(self, groups):
        return sum(len(g) for g in groups.values())


class CoverData(GObject.GObject):
    """Structured data for results from cover searching"""

    def __init__(self, url, source=None, dimensions=None):
        super().__init__()
        self.url = url
        self.dimensions = dimensions
        self.source = source

    def __repr__(self):
        return f"CoverData<url={self.url} @ {self.dimensions}>"



# Copyright 2023 LoveIsGrief
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
from typing import Any

from quodlibet import _, print_d, print_w
from quodlibet.formats import AudioFile
from quodlibet.plugins.query import QueryPlugin, QueryPluginError


class UniqueQuery(QueryPlugin):
    PLUGIN_ID = "unique_query"
    PLUGIN_NAME = _("Unique Query")
    PLUGIN_DESC = _("Filter search results by unique tags.")
    key = "unique"
    query_syntax = _("@(unique: tag)")
    query_description = (
        "<tt>tag</tt> can be album, artist, title or any other tag. "
        "Use multiple <tt>@(unique: tag)</tt> to filter by multiple tags."
    )

    usage = f"{query_syntax}\n\n{query_description}"

    def __init__(self):
        self.unique_tag_values = set()
        """The unique tag values that have been seen in the songs being filtered"""
        self._reported = set()
        """Unique errors to counter error log spam"""

    def search(self, song: AudioFile, body: Any | None) -> bool:
        return_value = False
        try:
            field_value = song[body]
            return_value = field_value not in self.unique_tag_values
            self.unique_tag_values.add(field_value)
            print_d(f"unique filtering value '{field_value}': {return_value}")
        except KeyError:
            pass
        except Exception as e:
            err_str = str(e)
            if err_str not in self._reported:
                self._reported.add(err_str)
                print_w(
                    f"{type(e).__name__} while filtering unique values for "
                    f"'{body}': {err_str}"
                )
        return return_value

    def parse_body(self, body: str) -> str:
        self.unique_tag_values.clear()
        if body is None:
            raise QueryPluginError
        unique_tag = body.strip()
        print_d(f"unique filtering tag: {unique_tag}")
        self._reported.clear()
        return unique_tag



# discord_status: Set Discord status as current song.
#
# Copyright (c) 2022 Aditi K <105543244+teeleafs@users.noreply.github.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _, app, print_e
from quodlibet.plugins import PluginConfig, ConfProp
from quodlibet.plugins.events import EventPlugin
from quodlibet.pattern import Pattern

from gi.repository import Gtk

try:
    from pypresence import Presence, InvalidID, DiscordNotFound
except ImportError as e:
    from quodlibet.plugins import MissingModulePluginError

    raise MissingModulePluginError("pypresence") from e

# The below resources are from/uploaded-to the Discord Application portal.
QL_DISCORD_RP_ID = "974521025356242984"
QL_LARGE_IMAGE = "io-github-quodlibet-quodlibet"

VERSION = "1.0"

# Default Rich Presence status lines.
CONFIG_DEFAULT_RP_LINE1 = "<artist> / <title>"
CONFIG_DEFAULT_RP_LINE2 = "<album>"


class DiscordStatusConfig:
    _config = PluginConfig(__name__)

    rp_line1 = ConfProp(_config, "rp_line1", CONFIG_DEFAULT_RP_LINE1)
    rp_line2 = ConfProp(_config, "rp_line2", CONFIG_DEFAULT_RP_LINE2)


discord_status_config = DiscordStatusConfig()


class DiscordStatusMessage(EventPlugin):
    PLUGIN_ID = _("Discord status message")
    PLUGIN_NAME = _("Discord Status Message")
    PLUGIN_DESC = _(
        "Change your Discord status message according to what "
        "you're currently listening to."
    )
    VERSION = VERSION

    def __init__(self):
        self.song = None
        self.discordrp = None

    def update_discordrp(self, details, state=None):
        if not self.discordrp:
            try:
                self.discordrp = Presence(QL_DISCORD_RP_ID, pipe=0)
                self.discordrp.connect()
            except (DiscordNotFound, ConnectionRefusedError) as e:
                print_e(f"Got error connecting to Discord ({e!r})")
                self.discordrp = None

        if self.discordrp:
            try:
                self.discordrp.update(
                    details=details, state=state, large_image=QL_LARGE_IMAGE
                )
            except InvalidID:
                # XXX Discord was closed?
                self.discordrp = None

    def handle_play(self):
        if self.song:
            details = Pattern(discord_status_config.rp_line1) % self.song
            state = Pattern(discord_status_config.rp_line2) % self.song

            # The details and state fields must be atleast 2 characters.
            if len(details) < 2:
                details = None

            if len(state) < 2:
                state = None

            self.update_discordrp(details, state)

    def handle_paused(self):
        self.update_discordrp(details=_("Paused"))

    def handle_unpaused(self):
        if not self.song:
            self.song = app.player.song
        self.handle_play()

    def plugin_on_song_started(self, song):
        self.song = song
        if not app.player.paused:
            self.handle_play()

    def plugin_on_paused(self):
        self.handle_paused()

    def plugin_on_unpaused(self):
        self.handle_unpaused()

    def enabled(self):
        if app.player.paused:
            self.handle_paused()
        else:
            self.handle_unpaused()

    def disabled(self):
        if self.discordrp:
            self.discordrp.clear()
            self.discordrp.close()
            self.discordrp = None
            self.song = None

    def PluginPreferences(self, parent):
        vb = Gtk.VBox(spacing=6)

        def rp_line1_changed(entry):
            discord_status_config.rp_line1 = entry.get_text()
            if not app.player.paused:
                self.handle_play()

        def rp_line2_changed(entry):
            discord_status_config.rp_line2 = entry.get_text()
            if not app.player.paused:
                self.handle_play()

        status_line1_box = Gtk.HBox(spacing=6)
        status_line1_box.set_border_width(3)

        status_line1 = Gtk.Entry()
        status_line1.set_text(discord_status_config.rp_line1)
        status_line1.connect("changed", rp_line1_changed)

        status_line1_box.pack_start(
            Gtk.Label(label=_("Status Line #1")), False, True, 0
        )
        status_line1_box.pack_start(status_line1, True, True, 0)

        status_line2_box = Gtk.HBox(spacing=3)
        status_line2_box.set_border_width(3)

        status_line2 = Gtk.Entry()
        status_line2.set_text(discord_status_config.rp_line2)
        status_line2.connect("changed", rp_line2_changed)

        status_line2_box.pack_start(
            Gtk.Label(label=_("Status Line #2")), False, True, 0
        )
        status_line2_box.pack_start(status_line2, True, True, 0)

        vb.pack_start(status_line1_box, True, True, 0)
        vb.pack_start(status_line2_box, True, True, 0)

        return vb



# Copyright 2013 Simonas Kazlauskas
#      2016-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gio, GLib, Soup

from quodlibet.plugins.cover import CoverSourcePlugin
from quodlibet.util.http import HTTPRequest, download_json
from quodlibet.util import print_w


class HTTPDownloadMixin:
    def download(self, message):
        request = HTTPRequest(message, self.cancellable)
        request.connect("sent", self._download_sent)
        request.connect("received", self._download_received)
        request.connect("failure", self._download_failure)
        request.send()

    def _download_sent(self, request, message):
        status = message.get_property("status-code")
        if not 200 <= status < 400:
            request.cancel()
            return self.fail(f"Bad HTTP code {status}")

        target = Gio.file_new_for_path(self.cover_path)
        flags = Gio.FileCreateFlags.NONE

        def replaced(cover_file, task, data):
            try:
                ostr = cover_file.replace_finish(task)
                request.provide_target(ostr)
                request.connect("receive-failure", self._receive_fail, target)
                request.receive()
            except GLib.GError:
                request.cancel()
                return self.fail("Cannot open cover file")

        target.replace_async(
            None, True, flags, GLib.PRIORITY_DEFAULT, self.cancellable, replaced, None
        )
        return None

    def _download_received(self, request, ostream):
        try:
            ostream.close(None)
        except GLib.GError as e:
            print_w(f"Got {e!r} trying to close handle after download")
        self.emit("fetch-success", self.cover)

    def _receive_fail(self, request, exception, gfile):
        def deleted(gfile, task, data):
            try:
                gfile.delete_finish(task)
            except GLib.GError:
                print_w("Could not clean up cover which failed to download")

        try:
            request.ostream.close(None)
        except GLib.GError as e:
            print_w(f"Got {e!r} trying to close handle during failure")
        gfile.delete_async(GLib.PRIORITY_DEFAULT, None, deleted, None)

    def _download_failure(self, request, exception):
        try:
            self.fail(exception.message or " ".join(exception.args))
        except AttributeError:
            self.fail(f"Download error ({exception})")


class ApiCoverSourcePlugin(CoverSourcePlugin, HTTPDownloadMixin):
    MIN_DIMENSION = 300
    """Minimum width / height in pixels for an image to be used"""

    @property
    def url(self):
        """The URL to the image, if remote"""
        return None

    def search(self):
        if not self.url:
            return self.emit("search-complete", [])
        msg = Soup.Message.new("GET", self.url)
        download_json(msg, self.cancellable, self._handle_search_response, None)
        return None

    def _handle_search_response(self, message, json_dict, data=None):
        self.emit("search-complete", [])

    def fetch_cover(self):
        if not self.url:
            return self.fail(f"Not enough data to get cover from {type(self).__name__}")

        def search_complete(self, res):
            self.disconnect(sci)
            if res:
                self.download(Soup.Message.new("GET", res[0]["cover"]))
            else:
                return self.fail("No cover was found")
            return None

        sci = self.connect("search-complete", search_complete)
        self.search()
        return None

    def _album_artists_for(self, song):
        """Returns a comma-separated list of artists indicating the
        "main" artists from the song's album"""
        people = [
            song.comma(key)
            for key in ["albumartist", "artist", "composer", "conductor", "performer"]
        ]
        people = list(filter(None, people))
        return people[0] if people else None


def escape_query_value(s):
    return GLib.Uri.escape_string(s, None, True)



# Copyright 2004-2006 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2012 Christoph Reiter
#      2013,2017 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _
from quodlibet import app
from quodlibet import qltk
from quodlibet.util import is_windows
from quodlibet.qltk import Icons
from quodlibet.pattern import Pattern
from quodlibet.qltk.entry import UndoEntry
from .util import pconfig


def supports_scrolling():
    """If our tray icon implementation supports scrolling"""

    return not is_windows()


class Preferences(Gtk.VBox):
    """A small window to configure the tray icon's tooltip."""

    def __init__(self):
        super().__init__(spacing=12)

        self.set_border_width(6)

        ccb = pconfig.ConfigCheckButton(
            _("Hide main window on close"), "window_hide", populate=True
        )
        self.pack_start(qltk.Frame(_("Behavior"), child=ccb), False, True, 0)

        def on_scroll_changed(button, new_state):
            if button.get_active():
                pconfig.set("modifier_swap", new_state)

        modifier_swap = pconfig.getboolean("modifier_swap")

        scrollwheel_box = Gtk.VBox(spacing=0)
        group = Gtk.RadioButton(
            group=None, label=_("Scroll wheel adjusts volume"), use_underline=True
        )
        group.connect("toggled", on_scroll_changed, False)
        group.set_active(not modifier_swap)
        scrollwheel_box.pack_start(group, False, True, 0)
        group = Gtk.RadioButton(
            group=group, label=_("Scroll wheel changes song"), use_underline=True
        )
        group.connect("toggled", on_scroll_changed, True)
        group.set_active(modifier_swap)
        scrollwheel_box.pack_start(group, False, True, 0)

        if supports_scrolling():
            self.pack_start(
                qltk.Frame(_("Scroll _Wheel"), child=scrollwheel_box), True, True, 0
            )

        box = Gtk.VBox(spacing=6)

        entry_box = Gtk.HBox(spacing=6)

        entry = UndoEntry()
        entry_box.pack_start(entry, True, True, 0)

        def on_reverted(*args):
            pconfig.reset("tooltip")
            entry.set_text(pconfig.gettext("tooltip"))

        revert = Gtk.Button()
        revert.add(
            Gtk.Image.new_from_icon_name(Icons.DOCUMENT_REVERT, Gtk.IconSize.BUTTON)
        )
        revert.connect("clicked", on_reverted)
        entry_box.pack_start(revert, False, True, 0)

        box.pack_start(entry_box, False, True, 0)

        preview = Gtk.Label()
        preview.set_line_wrap(True)
        preview_frame = Gtk.Frame(label=_("Preview"))
        vbox = Gtk.VBox(margin=18)
        vbox.pack_start(preview, False, False, 0)
        preview_frame.add(vbox)
        box.pack_start(preview_frame, False, True, 0)

        tt_frame = qltk.Frame(_("Tooltip Display"), child=box)
        tt_frame.get_label_widget().set_mnemonic_widget(entry)
        self.pack_start(tt_frame, True, True, 0)

        entry.connect("changed", self.__changed_entry, preview, preview_frame)
        entry.set_text(pconfig.gettext("tooltip"))

        for child in self.get_children():
            child.show_all()

    def __changed_entry(self, entry, label, frame):
        text = entry.get_text()

        if app.player.info is None:
            text = _("Not playing")
        else:
            text = Pattern(text) % app.player.info

        label.set_text(text)
        frame.set_tooltip_text(text)
        pconfig.set("tooltip", entry.get_text())
